#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from gi.repository import GLib  # @UnresolvedImport

from xpra.util.objects import typedict, AdHocStruct
from xpra.server.source.stub_source_mixin import StubSourceMixin


class ServerMixinTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super(ServerMixinTest, cls).setUpClass()
        cls.glib = GLib
        cls.main_loop = cls.glib.MainLoop()

    def setUp(self):
        self.mixin = None
        self.source = None
        self.protocol = None
        self.packet_handlers = {}
        self.legacy_alias = {}

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        if self.source:
            self.source.cleanup()
            self.source = None
        if self.mixin:
            self.mixin.cleanup()
            self.mixin = None

    def debug_all(self) -> None:
        from xpra.log import enable_debug_for
        enable_debug_for("all")

    def stop(self) -> None:
        self.glib.timeout_add(1000, self.main_loop.quit)

    def wait_for_threaded_init(self):
        # we don't do threading yet,
        # so no need to wait
        pass

    def add_packets(self, *packet_types: str, main_thread=True) -> None:
        for packet_type in packet_types:
            handler = getattr(self.mixin, "_process_" + packet_type.replace("-", "_"))
            self.add_packet_handler(packet_type, handler, main_thread)

    def add_legacy_alias(self, legacy_name: str, name: str) -> None:
        self.legacy_alias[legacy_name] = name

    def add_packet_handler(self, packet_type: str, handler=None, main_thread=True) -> None:
        self.packet_handlers[packet_type] = handler

    def handle_packet(self, packet):
        packet_type = packet[0]
        packet_type = self.legacy_alias.get(packet_type, packet_type)
        ph = self.packet_handlers.get(packet_type)
        assert ph is not None, "no packet handler for %s" % packet_type
        ph(self.protocol, packet)

    def verify_packet_error(self, packet):
        try:
            self.handle_packet(packet)
        except Exception:
            pass
        else:
            raise Exception("invalid packet %s should cause an error" % (packet,))

    def get_server_source(self, proto):
        assert proto==self.protocol
        return self.source

    def create_test_sockets(self):
        return {}

    def _test_mixin_class(self, mclass, opts, caps=None, source_mixin_class=StubSourceMixin):
        x = self.mixin = mclass()
        x._server_sources = {}   # pylint: disable=protected-access
        x.wait_for_threaded_init = self.wait_for_threaded_init
        x.add_packets = self.add_packets
        x.add_legacy_alias = self.add_legacy_alias
        x.add_packet_handler = self.add_packet_handler
        x.get_server_source = self.get_server_source
        x.init_state()
        x.init(opts)
        x.auth_classes = {}
        x.init_sockets(self.create_test_sockets())
        x.setup()
        x.threaded_setup()
        x.init_packet_handlers()
        caps = typedict(caps or {})
        send_ui = True
        self.source = None
        if source_mixin_class:
            self.source = source_mixin_class()
            self.protocol = AdHocStruct()
            self.protocol.TYPE = "xpra"
            self.source.wants = ("display", "foo")
            self.source.protocol = self.protocol
            self.source.init_from(self.protocol, x)
            self.source.init_state()
            self.source.parse_client_caps(caps)
            self.source.get_info()
        x.get_caps(self.source)
        x.get_info(None)
        x.parse_hello(self.source, caps, send_ui)
        x.get_info(self.source)



#!/usr/bin/env python3
#
# gtkPopupNotify.py
#
# Copyright (C) 2009 Daniel Woodhouse
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from xpra.os_util import OSX, gi_import
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label, modify_fg, color_parse
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.notifications.notifier_base import NotifierBase, log, NID

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")
GdkPixbuf = gi_import("GdkPixbuf")

DEFAULT_FG_COLOUR = None
DEFAULT_BG_COLOUR = None
if OSX:
    # black on white fits better with osx
    DEFAULT_FG_COLOUR = color_parse("black")
    DEFAULT_BG_COLOUR = color_parse("#f2f2f2")
DEFAULT_WIDTH = 340
DEFAULT_HEIGHT = 100


class GTKNotifier(NotifierBase):

    def __init__(self, closed_cb=None, action_cb=None, size_x=DEFAULT_WIDTH, size_y=DEFAULT_HEIGHT, timeout=5):
        super().__init__(closed_cb, action_cb)
        self.handles_actions = True
        """
        Create a new notification stack.  The recommended way to create Popup instances.
          Parameters:
            `size_x` : The desired width of the notifications.
            `size_y` : The desired minimum height of the notifications. If the text is
            longer it will be expanded to fit.
            `timeout` : Popup instance will disappear after this timeout if there
            is no human intervention. This can be overridden temporarily by passing
            a new timeout to the new_popup method.
        """
        self.size_x = size_x
        self.size_y = size_y
        self.timeout = timeout
        """
        Other parameters:
        These will take effect for every popup created after the change.
            `max_popups` : The maximum number of popups to be shown on the screen
            at one time.
            `bg_color` : if None default is used (usually grey). set with a gdk.Color.
            `fg_color` : if None default is used (usually black). set with a gdk.Color.
            `show_timeout : if True, a countdown till destruction will be displayed.

        """
        self.max_popups = 5
        self.fg_color = DEFAULT_FG_COLOUR
        self.bg_color = DEFAULT_BG_COLOUR
        self.show_timeout = False

        self._notify_stack = []
        self._offset = 0

        display = Gdk.Display.get_default()
        n = display.get_n_monitors()
        log("monitors=%s", n)
        # if n<2:
        monitor = display.get_monitor(0)
        geom = monitor.get_geometry()
        self.max_width = geom.width
        self.max_height = geom.height
        log("first monitor dimensions: %dx%d", self.max_width, self.max_height)
        self.x = self.max_width - 20  # keep away from the edge
        self.y = self.max_height - 64  # space for a panel
        log("our reduced dimensions: %dx%d", self.x, self.y)

    def cleanup(self):
        popups = tuple(self._notify_stack)
        self._notify_stack = []
        for x in popups:
            x.hide_notification()
        super().cleanup()

    def get_origin_x(self):
        return self.x

    def get_origin_y(self):
        return self.y

    def close_notify(self, nid):
        for x in self._notify_stack:
            if x.nid == nid:
                x.hide_notification()

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID, app_icon,
                    summary: str, body: str, actions, hints, timeout, icon):
        GLib.idle_add(self.new_popup, int(nid), summary, body, actions, icon, timeout, 0 < timeout <= 600)

    def new_popup(self, nid: int, summary: str, body: str, actions: tuple, icon, timeout=10 * 1000, show_timeout=False):
        """Create a new Popup instance, or update an existing one """
        existing = [p for p in self._notify_stack if p.nid == nid]
        if existing:
            existing[0].set_content(summary, body, actions, icon)
            return
        if len(self._notify_stack) == self.max_popups:
            oldest = self._notify_stack[0]
            oldest.hide_notification()
            self.popup_closed(oldest.nid, 4)
        image = None
        if icon and icon[0] == "png":
            img_data = icon[3]
            loader = GdkPixbuf.PixbufLoader()
            loader.write(img_data)
            loader.close()
            image = loader.get_pixbuf()
        popup = Popup(self, nid, summary, body, actions, image=image, timeout=timeout // 1000,
                      show_timeout=show_timeout)
        self._notify_stack.append(popup)
        self._offset += self._notify_stack[-1].h
        return False

    def destroy_popup_cb(self, popup):
        if popup in self._notify_stack:
            self._notify_stack.remove(popup)
            # move popups down if required
            offset = 0
            for note in self._notify_stack:
                offset = note.reposition(offset, self)
            self._offset = offset

    def popup_closed(self, nid, reason, text=""):
        if self.closed_cb:
            self.closed_cb(nid, reason, text)

    def popup_action(self, nid, action_id):
        if self.action_cb:
            self.action_cb(nid, action_id)


class Popup(Gtk.Window):
    def __init__(self, stack, nid, title, message, actions, image, timeout=5, show_timeout=False):
        log("Popup%s", (stack, nid, title, message, actions, image, timeout, show_timeout))
        self.stack = stack
        self.nid = nid
        super().__init__()

        self.set_accept_focus(False)
        self.set_focus_on_map(False)
        self.set_size_request(stack.size_x, -1)
        self.set_decorated(False)
        self.set_deletable(False)
        self.set_property("skip-pager-hint", True)
        self.set_property("skip-taskbar-hint", True)
        self.connect("enter-notify-event", self.on_hover, True)
        self.connect("leave-notify-event", self.on_hover, False)
        self.set_opacity(0.2)
        self.set_keep_above(True)
        self.destroy_cb = stack.destroy_popup_cb
        self.popup_closed = stack.popup_closed
        self.action_cb = stack.popup_action

        main_box = Gtk.VBox()
        header_box = Gtk.HBox()
        self.header = label()
        self.header.set_padding(3, 3)
        self.header.set_alignment(0, 0)
        header_box.pack_start(self.header, True, True, 5)
        icon = get_icon_pixbuf("close.png")
        if icon:
            close_button = Gtk.Image()
            close_button.set_from_pixbuf(icon)
            close_button.set_padding(3, 3)
            close_window = Gtk.EventBox()
            close_window.set_visible_window(False)
            close_window.connect("button-press-event", self.user_closed)
            close_window.add(close_button)
            close_window.set_size_request(icon.get_width(), icon.get_height())
            header_box.pack_end(close_window, False, False, 0)
        main_box.pack_start(header_box)

        body_box = Gtk.HBox()
        self.image = Gtk.Image()
        self.image.set_size_request(70, 70)
        self.image.set_alignment(0, 0)
        body_box.pack_start(self.image, False, False, 5)
        self.message = label()
        self.message.set_max_width_chars(80)
        self.message.set_size_request(stack.size_x - 90, -1)
        self.message.set_line_wrap(True)
        self.message.set_alignment(0, 0)
        self.message.set_padding(5, 10)
        self.counter = label()
        self.counter.set_alignment(1, 1)
        self.counter.set_padding(3, 3)
        self.timeout = timeout

        body_box.pack_start(self.message, True, False, 5)
        body_box.pack_end(self.counter, False, False, 5)
        main_box.pack_start(body_box, False, False, 5)

        self.buttons_box = Gtk.HBox(homogeneous=True)
        alignment = Gtk.Alignment(xalign=1.0, yalign=0.5, xscale=0.0, yscale=0.0)
        alignment.add(self.buttons_box)
        main_box.pack_start(alignment)
        self.add(main_box)
        if stack.bg_color is not None:
            self.modify_bg(Gtk.StateType.NORMAL, stack.bg_color)
        if stack.fg_color is not None:
            for widget in (self.message, self.header, self.counter):
                modify_fg(widget, stack.fg_color)
        self.show_timeout = show_timeout
        self.hover = False
        self.show_all()
        self.w = self.get_preferred_width()[0]
        self.h = self.get_preferred_height()[0]
        self.move(self.get_x(self.w), self.get_y(self.h))
        self.wait_timer = 0
        self.fade_out_timer = 0
        self.fade_in_timer = GLib.timeout_add(100, self.fade_in)
        # populate the window:
        self.set_content(title, message, actions, image)
        # ensure we don't show it in the taskbar:
        self.realize()
        self.get_window().set_skip_taskbar_hint(True)
        self.get_window().set_skip_pager_hint(True)
        add_close_accel(self, self.user_closed)

    def set_content(self, title, message, actions=(), image=None):
        self.header.set_markup("<b>%s</b>" % title)
        self.message.set_text(message)
        # remove any existing actions:
        for w in tuple(self.buttons_box.get_children()):
            self.buttons_box.remove(w)
        while len(actions) >= 2:
            action_id, action_text = actions[:2]
            actions = actions[2:]
            button = self.action_button(action_id, action_text)
            self.buttons_box.add(button)
        self.buttons_box.show_all()
        if image:
            self.image.show()
            self.image.set_from_pixbuf(image)
        else:
            self.image.hide()

    def action_button(self, action_id, action_text):
        button = Gtk.Button(label=action_text)
        button.set_relief(Gtk.ReliefStyle.NORMAL)

        def popup_cb_clicked(*args):
            self.hide_notification()
            log("popup_cb_clicked%s for action_id=%s, action_text=%s", args, action_id, action_text)
            self.action_cb(self.nid, action_id)

        button.connect("clicked", popup_cb_clicked)
        return button

    def get_x(self, w):
        x = self.stack.get_origin_x() - w // 2
        if (x + w) >= self.stack.max_width:  # don't overflow on the right
            x = self.stack.max_width - w
        x = max(0, x)  # or on the left
        log("get_x(%s)=%s", w, x)
        return x

    def get_y(self, h):
        y = self.stack.get_origin_y()
        if y >= (self.stack.max_height // 2):  # if near bottom, subtract window height
            y = y - h
        if (y + h) >= self.stack.max_height:
            y = self.stack.max_height - h
        y = max(0, y)
        log("get_y(%s)=%s", h, y)
        return y

    def reposition(self, offset, stack):
        """Move the notification window down, when an older notification is removed"""
        log("reposition(%s, %s)", offset, stack)
        new_offset = self.h + offset
        GLib.idle_add(self.move, self.get_x(self.w), self.get_y(new_offset))
        return new_offset

    def fade_in(self):
        opacity = self.get_opacity()
        opacity += 0.15
        if opacity >= 1:
            self.wait_timer = GLib.timeout_add(1000, self.wait)
            self.fade_in_timer = 0
            return False
        self.set_opacity(opacity)
        return True

    def wait(self):
        if not self.hover:
            self.timeout -= 1
        if self.show_timeout:
            self.counter.set_markup(str("<b>%s</b>" % max(0, self.timeout)))
        if self.timeout <= 0:
            self.fade_out_timer = GLib.timeout_add(100, self.fade_out)
            self.wait_timer = 0
            return False
        return True

    def fade_out(self):
        opacity = self.get_opacity()
        opacity -= 0.10
        if opacity <= 0:
            self.in_progress = False
            self.hide_notification()
            self.fade_out_timer = 0  # redundant
            self.popup_closed(self.nid, 1)
            return False
        self.set_opacity(opacity)
        return True

    def on_hover(self, _window, _event, hover):
        """Starts/Stops the notification timer on a mouse in/out event"""
        self.hover = hover

    def user_closed(self, *_args):
        self.hide_notification()
        self.popup_closed(self.nid, 2)

    def hide_notification(self):
        """Destroys the notification and tells the stack to move the
        remaining notification windows"""
        log("hide_notification()")
        for timer in ("fade_in_timer", "fade_out_timer", "wait_timer"):
            v = getattr(self, timer)
            if v:
                setattr(self, timer, None)
                GLib.source_remove(v)
        # destroy window from the UI thread:
        GLib.idle_add(self.destroy)
        self.close()
        self.destroy_cb(self)


def main():
    # example usage
    import random
    color_combos = (("red", "white"), ("white", "blue"), ("green", "black"))
    messages: list[tuple[int, str, str, tuple]] = [
        (1, "Hello", "This is a popup", ()),
        (2, "Actions", "This notification has 3 actions", (1, "Action 1", 2, "Action 2", 3, "Action 3")),
        (3, "Some Latin", "Quidquid latine dictum sit, altum sonatur.", ()),
        (4, "A long message", "The quick brown fox jumped over the lazy dog. " * 6, ()),
        (1, "Hello Again", "Replacing the first notification", ()),
        (2, "Actions Again", "Replacing with just 1 action", (999, "Action 999")),
    ]

    # images = ("logo1_64.png", None)

    def notify_factory():
        color = random.choice(color_combos)
        nid, title, message, actions = messages.pop(0)
        icon = ()  # random.choice(images)
        notifier.bg_color = color_parse(color[0])
        notifier.fg_color = color_parse(color[1])
        notifier.show_timeout = random.choice((True, False))
        notifier.new_popup(nid, title, message, actions, icon)
        return len(messages)

    def gtk_main_quit():
        print("quitting")
        Gtk.main_quit()

    notifier = GTKNotifier(timeout=6)
    GLib.timeout_add(4000, notify_factory)
    GLib.timeout_add(30000, gtk_main_quit)
    Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")


def make_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_title("Text Entry")
    window.connect("destroy", Gtk.main_quit)
    window.set_default_size(320, -1)
    window.set_border_width(40)
    window.set_position(Gtk.WindowPosition.CENTER)
    icon = get_icon_pixbuf("font.png")
    if icon:
        window.set_icon(icon)

    vbox = Gtk.VBox(homogeneous=False, spacing=10)
    window.add(vbox)
    entry = Gtk.Entry()
    entry.set_text("hello")
    vbox.add(entry)

    textview = Gtk.TextView()
    textbuffer = textview.get_buffer()
    textbuffer.set_text("Sample text\nmultiline")
    textview.set_editable(True)
    textview.set_size_request(200, 80)
    vbox.add(textview)

    window.add(entry)
    return window


def main():
    with program_context("text-entry", "Text Entry"):
        w = make_window()
        add_close_accel(w, Gtk.main_quit)

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        GLib.idle_add(show_with_focus)
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("text entry test window")
        Gtk.main()


if __name__ == "__main__":
    main()
    sys.exit(0)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

import os.path
import hashlib
from typing import Any
from collections.abc import Sequence

from xpra.util.stats import to_std_unit, std_unit
from xpra.os_util import WIN32, POSIX
from xpra.util.env import osexpand
from xpra.util.io import load_binary_file
from xpra.util.str_fn import repr_ellipsized, csv
from xpra.common import NotificationID
from xpra.auth.auth_helper import AuthDef
from xpra.net.common import PacketType
from xpra.net.file_transfer import FileTransferAttributes
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

printlog = Logger("printing")
filelog = Logger("file")

SAVE_PRINT_JOBS = os.environ.get("XPRA_SAVE_PRINT_JOBS", None)


def _save_print_job(filename, file_data) -> None:
    save_filename = os.path.join(SAVE_PRINT_JOBS, filename)
    try:
        with open(save_filename, "wb") as f:
            f.write(file_data)
        printlog.info("saved print job to: %s", save_filename)
    except Exception as e:
        printlog.error("Error: failed to save print job to %s", save_filename)
        printlog.estr(e)


class FilePrintServer(StubServerMixin):
    """
    Mixin for servers that can handle file transfers and forwarded printers.
    Printer forwarding is only supported on Posix servers with the cups backend script.
    """

    def __init__(self):
        self.lpadmin: str = ""
        self.lpinfo: str = ""
        self.add_printer_options = []
        self.file_transfer = FileTransferAttributes()

    def init(self, opts) -> None:
        self.file_transfer.init_opts(opts, can_ask=False)
        self.lpadmin = opts.lpadmin
        self.lpinfo = opts.lpinfo
        self.add_printer_options = opts.add_printer_options
        # server-side printer handling is only for posix via pycups for now:
        self.postscript_printer = opts.postscript_printer
        self.pdf_printer = opts.pdf_printer

    def threaded_setup(self) -> None:
        self.init_printing()

    def init_sockets(self, sockets) -> None:
        # verify we have a local socket for printing:
        unixsockets = [info for socktype, _, info, _ in sockets if socktype == "socket"]
        printlog("local unix domain sockets we can use for printing: %s", unixsockets)
        if not unixsockets and self.file_transfer.printing:
            if not WIN32:
                printlog.warn("Warning: no local sockets defined,")
                printlog.warn(" disabling printer forwarding")
            printlog("printer forwarding disabled")
            self.file_transfer.printing = False

    def get_server_features(self, _source) -> dict[str, Any]:
        f = self.file_transfer.get_file_transfer_features()
        f["printer.attributes"] = ("printer-info", "device-uri")
        ftf = self.file_transfer.get_file_transfer_features()
        if self.file_transfer.file_transfer:
            ftf["request-file"] = True
        f.update(ftf)
        return f

    def get_info(self, _proto) -> dict[str, Any]:
        d = {}
        if POSIX:
            d.update({
                "lpadmin": self.lpadmin,
                "lpinfo": self.lpinfo,
                "add-printer-options": self.add_printer_options,
            })
        if self.file_transfer.printing:
            from xpra.platform.printing import get_info
            d.update(get_info())
        info = {"printing": d}
        if self.file_transfer.file_transfer:
            fti = self.file_transfer.get_info()
            if self.file_transfer.file_transfer:
                fti["request-file"] = True
            info["file"] = fti
        return info

    def init_printing(self) -> None:
        printing = self.file_transfer.printing
        if not printing or WIN32:
            return
        try:
            from xpra.platform import pycups_printing
            pycups_printing.set_lpadmin_command(self.lpadmin)
            pycups_printing.set_lpinfo_command(self.lpinfo)
            pycups_printing.set_add_printer_options(self.add_printer_options)
            if self.postscript_printer:
                pycups_printing.add_printer_def("application/postscript", self.postscript_printer)
            if self.pdf_printer:
                pycups_printing.add_printer_def("application/pdf", self.pdf_printer)
            printer_definitions = pycups_printing.get_printer_definitions()
            printing = bool(printer_definitions)
            if printing:
                printlog.info("printer forwarding enabled using %s", " and ".join(
                    x.replace("application/", "") for x in printer_definitions))
            else:
                printlog.warn("Warning: no printer definitions found,")
                printlog.warn(" cannot enable printer forwarding")
        except ImportError as e:
            printlog("printing module is not installed: %s", e)
            printing = False
        except Exception:
            printlog.error("Error: failed to set lpadmin and lpinfo commands", exc_info=True)
            printing = False
        # verify that we can talk to the socket:
        auth_classes: Sequence[AuthDef] = self.auth_classes.get("socket", ())
        if printing and auth_classes:
            fail: list[str] = []
            for auth_class in auth_classes:
                try:
                    # this should be the name of the auth module:
                    auth_name = auth_class[0]
                except Exception:
                    auth_name = str(auth_class)
                if auth_name not in ("allow", "file", "hosts", "none", "peercred"):
                    fail.append(auth_name)
            if fail:
                printlog.warn("Warning: printer forwarding is not supported")
                printlog.warn(" with sockets using authentication modules %s", csv(fail))
                printing = False
        # update file transfer attributes since printing nay have been disabled here
        self.file_transfer.printing = printing
        printlog("init_printing() printing=%s", printing)

    def _process_print(self, _proto, packet: PacketType) -> None:
        # ie: from the xpraforwarder we call this command:
        # command = ["xpra", "print", "socket:/path/tosocket",
        #           filename, mimetype, source, title, printer, no_copies, print_options]
        assert self.file_transfer.printing
        # printlog("_process_print(%s, %s)", proto, packet)
        if len(packet) < 3:
            printlog.error("Error: invalid print packet, only %i arguments", len(packet))
            printlog.error(" %s", [repr_ellipsized(x) for x in packet])
            return
        filename = str(packet[1])
        file_data = packet[2]
        mimetype, source_uuid, title, printer, no_copies, print_options = "", "*", "unnamed document", "", 1, ""
        if len(packet) >= 4:
            mimetype = str(packet[3])
        if len(packet) >= 5:
            source_uuid = str(packet[4])
        if len(packet) >= 6:
            title = str(packet[5])
        if len(packet) >= 7:
            printer = str(packet[6])
        if len(packet) >= 8:
            no_copies = int(packet[7])
        if len(packet) >= 9:
            print_options = packet[8]
        # parse and validate:
        if len(mimetype) >= 128:
            printlog.error("Error: invalid mimetype in print packet:")
            printlog.error(" %s", repr_ellipsized(mimetype))
            return
        if not isinstance(print_options, dict):
            s = str(print_options)
            print_options = {}
            for x in s.split(" "):
                parts = x.split("=", 1)
                if len(parts) == 2:
                    print_options[parts[0]] = parts[1]
        printlog("process_print: %s", (filename, mimetype, "%s bytes" % len(file_data),
                                       source_uuid, title, printer, no_copies, print_options))
        printlog("process_print: got %s bytes for file %s", len(file_data), filename)
        # parse the print options:
        hu = hashlib.sha256()
        hu.update(file_data)
        printlog("sha1 digest: %s", hu.hexdigest())
        options = {
            "printer": printer,
            "title": title,
            "copies": no_copies,
            "options": print_options,
            "sha256": hu.hexdigest(),
        }
        printlog("parsed printer options: %s", options)
        if SAVE_PRINT_JOBS:
            _save_print_job(filename, file_data)

        sent = 0
        sources = tuple(self._server_sources.values())
        printlog("will try to send to %i clients: %s", len(sources), sources)
        for ss in sources:
            if source_uuid not in ("*", ss.uuid):
                printlog("not sending to %s (uuid=%s, wanted uuid=%s)", ss, ss.uuid, source_uuid)
                continue
            if not ss.printing:
                if source_uuid != '*':
                    printlog.warn("Warning: printing is not enabled for:")
                    printlog.warn(" %s", ss)
                else:
                    printlog("printing is not enabled for %s", ss)
                continue
            if not ss.printers:
                printlog.warn("Warning: client %s does not have any printers", ss.uuid)
                continue
            if printer not in ss.printers:
                printlog.warn("Warning: client %s does not have a '%s' printer", ss.uuid, printer)
                continue
            printlog("'%s' sent to %s for printing on '%s'", title or filename, ss, printer)
            if ss.send_file(filename, mimetype, file_data, len(file_data), True, True, options):
                sent += 1
        # warn if not sent:
        log_fn = printlog.warn if sent == 0 else printlog.info
        unit_str, v = to_std_unit(len(file_data), unit=1024)
        log_fn("'%s' (%i%sB) sent to %i clients for printing", title or filename, v, unit_str, sent)

    def _process_printers(self, proto, packet: PacketType) -> None:
        if not self.file_transfer.printing or WIN32:
            printlog.error("Error: received printer definitions data")
            printlog.error(" but this server does not support printer forwarding")
            return
        ss = self.get_server_source(proto)
        if ss is None:
            return
        printers = dict(packet[1])
        auth_class: Sequence[AuthDef] = self.auth_classes.get("socket", ())
        ss.set_printers(printers, self.password_file, auth_class, self.encryption, self.encryption_keyfile)

    ######################################################################
    # file transfers:
    def _process_send_file(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            printlog.warn("Warning: invalid client source for send-file packet")
            return
        ss._process_send_file(packet)

    def _process_ack_file_chunk(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            printlog.warn("Warning: invalid client source for ack-file-chunk packet")
            return
        ss._process_ack_file_chunk(packet)

    def _process_send_file_chunk(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            printlog.warn("Warning: invalid client source for send-file-chunk packet")
            return
        ss._process_send_file_chunk(packet)

    def _process_send_data_request(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            printlog.warn("Warning: invalid client source for send-file-request packet")
            return
        ss._process_send_data_request(packet)

    def _process_send_data_response(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            printlog.warn("Warning: invalid client source for send-data-response packet")
            return
        ss._process_send_data_response(packet)

    def _process_request_file(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            filelog.warn("Warning: invalid client source for send-data-response packet")
            return
        argf = str(packet[1])
        if argf == "${XPRA_SERVER_LOG}" and not os.environ.get("XPRA_SERVER_LOG"):
            filelog("no server log to send")
            return
        openit = packet[2]
        filename = os.path.abspath(osexpand(argf))
        if not os.path.exists(filename):
            filelog.warn("Warning: the file requested does not exist:")
            filelog.warn(f" {filename!r}")
            ss.may_notify(NotificationID.FILETRANSFER,
                          "File not found", "The file requested does not exist:\n%s" % filename,
                          icon_name="file")
            return
        try:
            stat = os.stat(filename)
            filelog("os.stat(%s)=%s", filename, stat)
        except os.error:
            filelog("os.stat(%s)", filename, exc_info=True)
        else:
            file_size = stat.st_size
            if file_size > self.file_transfer.file_size_limit or file_size > ss.file_size_limit:
                ss.may_notify(NotificationID.FILETRANSFER,
                              "File too large",
                              "The file requested is too large to send:\n%s\nis %s" % (argf, std_unit(file_size)),
                              icon_name="file")
                return
        data = load_binary_file(filename)
        ss.send_file(filename, "", data, len(data), openit=openit, options={"request-file": (argf, openit)})

    def init_packet_handlers(self) -> None:
        # noqa: E241
        if self.file_transfer.printing:
            self.add_packets("printers", "print")
        if self.file_transfer.printing or self.file_transfer.file_transfer:
            self.add_packets("send-file", "ack-file-chunk", "send-file-chunk",
                             "send-data-request", "send-data-response")
        if self.file_transfer.file_transfer:
            self.add_packets("request-file")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import uuid
import struct
from types import ModuleType
from typing import NoReturn

# only minimal imports go at the top
# so that this file can be included everywhere
# without too many side effects
# pylint: disable=import-outside-toplevel

WIN32: bool = sys.platform.startswith("win")
OSX: bool = sys.platform.startswith("darwin")
LINUX: bool = sys.platform.startswith("linux")
NETBSD: bool = sys.platform.startswith("netbsd")
OPENBSD: bool = sys.platform.startswith("openbsd")
FREEBSD: bool = sys.platform.startswith("freebsd")
POSIX: bool = os.name == "posix"
BITS: int = struct.calcsize(b"P")*8


GIR_VERSIONS: dict[str, str] = {
    "Gtk": "3.0",
    "Gdk": "3.0",
    "GdkX11": "3.0",
    "Pango": "1.0",
    "PangoCairo": "1.0",
    "GLib": "2.0",
    "GObject": "2.0",
    "GdkPixbuf": "2.0",
    "IBus": "1.0",
    "Gio": "2.0",
    "Rsvg": "2.0",
    "Gst": "1.0",
    "NM": "1.0",
    "GtkosxApplication": "1.0",
    "Notify": "0.7",
}


def gi_import(mod="Gtk", version="") -> ModuleType:
    version = version or GIR_VERSIONS.get(mod, "")
    from xpra.util.env import SilenceWarningsContext
    with SilenceWarningsContext(DeprecationWarning, ImportWarning):
        import gi
        try:
            gi.require_version(mod, version)
        except ValueError as e:
            raise ImportError(f"unable to import {mod!r} {version=!r}: {e}") from None
        import importlib
        return importlib.import_module(f"gi.repository.{mod}")


def is_admin() -> bool:
    if WIN32:
        from ctypes import windll
        return windll.shell32.IsUserAnAdmin() != 0
    return os.geteuid() == 0


def getuid() -> int:
    if POSIX:
        return os.getuid()
    return 0


def getgid() -> int:
    if POSIX:
        return os.getgid()
    return 0


def get_shell_for_uid(uid: int) -> str:
    if POSIX:
        from pwd import getpwuid
        try:
            return getpwuid(uid).pw_shell
        except KeyError:
            pass
    return ""


def get_username_for_uid(uid: int) -> str:
    if POSIX:
        from pwd import getpwuid
        try:
            return getpwuid(uid).pw_name
        except KeyError:
            pass
    return ""


def get_home_for_uid(uid: int) -> str:
    if POSIX:
        from pwd import getpwuid
        try:
            return getpwuid(uid).pw_dir
        except KeyError:
            pass
    return ""


def get_groups(username: str) -> list[str]:
    if POSIX:
        import grp
        return [gr.gr_name for gr in grp.getgrall() if username in gr.gr_mem]
    return []


def get_group_id(group: str) -> int:
    try:
        import grp
        gr = grp.getgrnam(group)
        return gr.gr_gid
    except (ImportError, KeyError):
        return -1


def get_hex_uuid() -> str:
    return uuid.uuid4().hex


def get_int_uuid() -> int:
    return uuid.uuid4().int


def get_machine_id() -> str:
    """
        Try to get uuid string which uniquely identifies this machine.
        Warning: only works on posix!
        (which is ok since we only used it on posix at present)
    """
    v = ""
    if POSIX:
        from xpra.util.io import load_binary_file
        for filename in ("/etc/machine-id", "/var/lib/dbus/machine-id"):
            b = load_binary_file(filename)
            if b:
                v = b.decode("latin1")
                break
    elif WIN32:
        v = str(uuid.getnode())
    return v.strip("\n\r")


def get_user_uuid() -> str:
    """
        Try to generate an uuid string which is unique to this user.
        (relies on get_machine_id to uniquely identify a machine)
    """
    user_uuid = os.environ.get("XPRA_USER_UUID")
    if user_uuid:
        return user_uuid
    import hashlib
    u = hashlib.sha256()

    def uupdate(ustr) -> None:
        u.update(ustr.encode("utf-8"))
    uupdate(get_machine_id())
    if POSIX:
        uupdate("/")
        uupdate(str(os.getuid()))
        uupdate("/")
        uupdate(str(os.getgid()))
    uupdate(os.path.expanduser("~/"))
    return u.hexdigest()


# here so we can override it when needed
def force_quit(status=1) -> NoReturn:
    # noinspection PyProtectedMember
    os._exit(int(status))  # pylint: disable=protected-access



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import sys
import logging
import weakref
import itertools
from typing import Any, Final
from collections.abc import Callable, Sequence
from contextlib import AbstractContextManager
# This module is used by non-GUI programs and thus must not import gtk.

LOG_PREFIX: str = ""
LOG_FORMAT: str = "%(asctime)s %(message)s"
DEBUG_MODULES: Sequence[str] = ()
if os.name != "posix" or os.getuid() != 0:
    LOG_FORMAT = os.environ.get("XPRA_LOG_FORMAT", LOG_FORMAT)
    LOG_PREFIX = os.environ.get("XPRA_LOG_PREFIX", LOG_PREFIX)
    DEBUG_MODULES = tuple(x.strip() for x in os.environ.get("XPRA_DEBUG_MODULES", "").split(",") if x.strip())
NOPREFIX_FORMAT: Final[str] = "%(message)s"


BACKTRACE_LEVEL = int(os.environ.get("XPRA_LOG_BACKTRACE_LEVEL", logging.CRITICAL))
BACKTRACE_REGEXES: Sequence[str] = tuple(
    x for x in os.environ.get("XPRA_LOG_BACKTRACE_REGEXES", "").split("|") if x
)


logging.basicConfig(format=LOG_FORMAT)
logging.root.setLevel(logging.INFO)

debug_enabled_categories: set[str] = set()
debug_disabled_categories: set[str] = set()
backtrace_expressions: set[re.Pattern] = set()


MODULE_FILE = os.path.join(os.sep, "xpra", "log.py")        # ie: "/xpra/log.py"


def get_debug_args() -> Sequence[str]:
    args = []
    if debug_enabled_categories:
        args += list(debug_enabled_categories)
    if debug_disabled_categories:
        for x in debug_disabled_categories:
            args.append(f"-{x}")
    return args


class FullDebugContext:
    __slots__ = ("debug_enabled_categories", "enabled")

    def __enter__(self):
        self.debug_enabled_categories = debug_enabled_categories
        debug_enabled_categories.clear()
        debug_enabled_categories.add("all")
        self.enabled = []
        for x in get_all_loggers():
            if not x.is_debug_enabled():
                self.enabled.append(x)
                x.enable_debug()

    def __exit__(self, *_args):
        for x in self.enabled:
            x.disable_debug()
        debug_enabled_categories.clear()
        debug_enabled_categories.update(self.debug_enabled_categories)


def add_debug_category(*cat) -> None:
    remove_disabled_category(*cat)
    for c in cat:
        debug_enabled_categories.add(c)


def remove_debug_category(*cat) -> None:
    for c in cat:
        if c in debug_enabled_categories:
            debug_enabled_categories.remove(c)


def is_debug_enabled(category: str) -> bool:
    if "all" in debug_enabled_categories:
        return True
    if category in debug_enabled_categories:
        return True
    return isenvdebug(category) or isenvdebug("ALL")


def add_disabled_category(*cat: str) -> None:
    remove_debug_category(*cat)
    for c in cat:
        debug_disabled_categories.add(c)


def remove_disabled_category(*cat: str) -> None:
    for c in cat:
        if c in debug_disabled_categories:
            debug_disabled_categories.remove(c)


def add_backtrace(*expressions: str) -> None:
    for e in expressions:
        backtrace_expressions.add(re.compile(e))


def remove_backtrace(*expressions: str) -> None:
    for e in expressions:
        try:
            backtrace_expressions.remove(re.compile(e))
        except KeyError:
            pass


add_backtrace(*BACKTRACE_REGEXES)


default_level: int = logging.DEBUG


def set_default_level(level: int) -> None:
    global default_level
    default_level = level


def standard_logging(log, level: int, msg: str, *args, **kwargs) -> None:
    # this function just delegates to the regular python stdlib logging `log`:
    kwargs.pop("remote", None)
    log(level, msg, *args, **kwargs)


# this allows us to capture all logging and redirect it:
# the default 'standard_logging' uses the logger,
# but the client may inject its own handler here
global_logging_handler: Callable = standard_logging


def set_global_logging_handler(h: Callable) -> Callable:
    assert callable(h)
    global global_logging_handler
    saved = global_logging_handler
    global_logging_handler = h
    return saved


def setloghandler(lh) -> None:
    logging.root.handlers = []
    logging.root.addHandler(lh)


def enable_color(to=sys.stdout, format_string=NOPREFIX_FORMAT) -> None:
    if not hasattr(to, "fileno") or not hasattr(to, "buffer"):
        # on win32 sys.stdout can be a "Blackhole",
        # which does not have a fileno
        return
    # pylint: disable=import-outside-toplevel
    import codecs
    to = codecs.getwriter("utf-8")(to.buffer, "replace")
    try:
        from xpra.util.colorstreamhandler import ColorStreamHandler
    except ImportError:
        pass
    else:
        csh = ColorStreamHandler(to)
        csh.setFormatter(logging.Formatter(format_string))
        setloghandler(csh)


def enable_format(format_string: str) -> None:
    try:
        logging.root.handlers[0].formatter = logging.Formatter(format_string)
    except (AttributeError, IndexError):
        pass


def consume_verbose_argv(argv: list[str], *categories: str) -> bool:
    verbose = False
    for x in list(argv):
        if x in ("-v", "--verbose"):
            verbose = True
            argv.remove(x)
            for category in categories:
                add_debug_category(category)
                enable_debug_for(category)
    return verbose


# noinspection PyPep8
STRUCT_KNOWN_FILTERS: dict[str, dict[str, str]] = {
    "Client": {
        "client"        : "All client code",
        "paint"         : "Client window paint code",
        "draw"          : "Client draw packets",
        "cairo"         : "Cairo paint code used with the GTK3 client",
        "opengl"        : "Client OpenGL rendering",
        "info"          : "About and Session info dialogs",
        "launcher"      : "The client launcher program",
    },
    "General": {
        "clipboard"     : "All clipboard operations",
        "notify"        : "Notification forwarding",
        "tray"          : "System Tray forwarding",
        "printing"      : "Printing",
        "file"          : "File transfers",
        "keyboard"      : "Keyboard mapping and key event handling",
        "ibus"          : "IBus keyboard layouts",
        "screen"        : "Screen and workarea dimension",
        "fps"           : "Frames per second",
        "xsettings"     : "XSettings synchronization",
        "dbus"          : "DBUS calls",
        "menu"          : "Menus",
        "events"        : "System and window events",
    },
    "Window": {
        "window"        : "All window code",
        "damage"        : "Window X11 repaint events",
        "geometry"      : "Window geometry",
        "shape"         : "Window shape forwarding (XShape)",
        "focus"         : "Window focus",
        "workspace"     : "Window workspace synchronization",
        "metadata"      : "Window metadata",
        "alpha"         : "Window Alpha channel (transparency)",
        "state"         : "Window state",
        "icon"          : "Window icons",
        "frame"         : "Window frame",
        "grab"          : "Window grabs (both keyboard and mouse)",
        "dragndrop"     : "Window drag-n-drop events",
        "filters"       : "Window filters",
    },
    "Encoding": {
        "codec"         : "Codec loader and video helper",
        "loader"        : "Pixel compression codec loader",
        "video"         : "Video encoding",
        "score"         : "Video pipeline scoring and selection",
        "encoding"      : "Server side encoding selection and compression",
        "scaling"       : "Picture scaling",
        "scroll"        : "Scrolling detection and compression",
        "xor"           : "XOR delta pre-compression",
        "subregion"     : "Video subregion processing",
        "regiondetect"  : "Video region detection",
        "regionrefresh" : "Video region refresh",
        "refresh"       : "Refresh of lossy screen updates",
        "compress"      : "Pixel compression",
    },
    "Codec": {
        "csc"           : "Colourspace conversion codecs",
        "cuda"          : "CUDA device access",
        "cython"        : "Cython CSC module",
        "libyuv"        : "libyuv CSC module",
        "decoder"       : "All decoders",
        "encoder"       : "All encoders",
        "argb"          : "ARGB encoder",
        "pillow"        : "Pillow encoder and decoder",
        "spng"          : "spng codec",
        "jpeg"          : "JPEG codec",
        "vpx"           : "libvpx encoder and decoder",
        "amf"           : "amf encoder",
        "nvjpeg"        : "nvidia nvjpeg hardware encoder",
        "nvenc"         : "nvidia nvenc video hardware encoder",
        "nvdec"         : "nvidia nvdec video hardware decoder",
        "nvfbc"         : "nvidia nvfbc screen capture",
        "x264"          : "libx264 encoder",
        "openh264"      : "openh264 decoder",
        "webp"          : "libwebp encoder and decoder",
        "avif"          : "libavif encoder and decoder",
        "webcam"        : "webcam access",
        "evdi"          : "evdi virtual monitor",
        "drm"           : "direct rendering manager",
        "remote"        : "remote server codec",
    },
    "Pointer": {
        "mouse"         : "Mouse motion",
        "cursor"        : "Mouse cursor shape",
    },
    "Misc": {
        # libraries
        "gtk"           : "All GTK code: bindings, client, etc",
        "util"          : "All utility functions",
        "gobject"       : "Command line clients",
        "brotli"        : "Brotli bindings",
        "lz4"           : "LZ4 bindings",
        # server bits:
        "test"          : "Test code",
        "verbose"       : "Very verbose flag",
        # specific applications:
    },
    "Network": {
        # internal / network:
        "network"       : "All network code",
        "bandwidth"     : "Bandwidth detection and management",
        "ssh"           : "SSH connections",
        "ssl"           : "SSL connections",
        "http"          : "HTTP requests",
        "rfb"           : "RFB Protocol",
        "mmap"          : "mmap transfers",
        "protocol"      : "Packet input and output (formatting, parsing, sending and receiving)",
        "websocket"     : "WebSocket layer",
        "named-pipe"    : "Named pipe",
        "crypto"        : "Encryption",
        "auth"          : "Authentication",
        "upnp"          : "UPnP",
        "quic"          : "QUIC",
        "ping"          : "ping",
    },
    "Server": {
        "server"        : "All server code",
        "proxy"         : "Proxy server",
        "shadow"        : "Shadow server",
        "command"       : "Server control channel",
        "timeout"       : "Server timeouts",
        "exec"          : "Executing commands",
        # server features:
        "mdns"          : "mDNS session publishing",
        # server internals:
        "stats"         : "Server statistics",
        "xshm"          : "XShm pixel capture",
    },
    "Audio": {
        "audio"         : "All audio",
        "gstreamer"     : "GStreamer internal messages",
        "av-sync"       : "Audio-video sync",
    },
    "X11": {
        "x11"           : "All X11 code",
        "xinput"        : "XInput bindings",
        "bindings"      : "X11 Cython bindings",
        "core"          : "X11 core bindings",
        "randr"         : "X11 RandR bindings",
        "record"        : "X11 Record bindings",
        "ximage"        : "X11 XImage bindings",
        "error"         : "X11 errors",
    },
    "Platform": {
        "platform"      : "All platform support code",
        "import"        : "Platform support import code",
        "osx"           : "Mac OS X platform support code",
        "win32"         : "Microsoft Windows platform support code",
        "d3d11"         : "Microsoft Direct 3D 11",
        "posix"         : "Posix platform code",
    },
}

# flatten it:
KNOWN_FILTERS: dict[str, str] = {}
for d in STRUCT_KNOWN_FILTERS.values():
    for k, v in d.items():
        KNOWN_FILTERS[k] = v


def isenvdebug(category: str) -> bool:
    return os.environ.get("XPRA_%s_DEBUG" % category.upper().replace("-", "_").replace("+", "_"), "0") == "1"


def get_info() -> dict[str, Any]:
    info = {
        "categories": {
            "enabled"   : tuple(debug_enabled_categories),
            "disabled"  : tuple(debug_disabled_categories),
        },
        "backtrace-level": BACKTRACE_LEVEL,
        "backtrace-expressions": tuple(bt.pattern for bt in backtrace_expressions),
        "handler"   : getattr(global_logging_handler, "__name__", "<unknown>"),
        "prefix"    : LOG_PREFIX,
        "format"    : LOG_FORMAT,
        "debug-modules" : DEBUG_MODULES,
    }
    from xpra.common import FULL_INFO
    if FULL_INFO > 1:
        info["filters"] = STRUCT_KNOWN_FILTERS
    return info


class Logger:
    """
    A wrapper around 'logging' with some convenience stuff.  In particular:
    * You initialize it with a list of categories
        If unset, the default logging target is set to the name of the module where
        Logger() was called.
    * Any of the categories can enable debug logging if the environment
    variable 'XPRA_${CATEGORY}_DEBUG' is set to "1"
    * We also keep a list of debug_categories, so these can get enabled
        programmatically too
    * We keep track of which loggers are associated with each category,
        so we can enable/disable debug logging by category
    * You can pass exc_info=True to any method, and sys.exc_info() will be
        substituted.
    * __call__ is an alias for debug
    * we bypass the logging system unless debugging is enabled for the logger,
        which is much faster than relying on the python logging code
    """
    __slots__ = ("categories",
                 "level", "level_override", "min_level", "_logger", "debug_enabled", "__weakref__", "debug")

    def __init__(self, *categories: str):
        self.debug = self.__call__
        self.min_level = 0
        self.categories = list(categories)
        n = 1
        caller = ""
        while n < 10:
            try:
                # noinspection PyProtectedMember
                caller = sys._getframe(n).f_globals["__name__"]  # pylint: disable=protected-access
                if caller == "__main__" or caller.startswith("importlib"):
                    n += 1
                else:
                    break
            except (AttributeError, ValueError):
                break
        if caller and caller != "__main__" and not caller.startswith("importlib"):
            self.categories.insert(0, caller)
        self.level = logging.INFO
        self.level_override = 0
        self._logger = logging.getLogger(".".join(self.categories))
        self.setLevel(default_level)
        disabled = False
        enabled = False
        if caller in DEBUG_MODULES:
            enabled = True
        else:
            for cat in self.categories:
                if cat in debug_disabled_categories:
                    disabled = True
                if is_debug_enabled(cat):
                    enabled = True
            if len(categories) > 1:
                # try all string permutations of those categories:
                # "keyboard", "events" -> "keyboard+events" or "events+keyboard"
                for cats in itertools.permutations(categories):
                    cstr = "+".join(cats)
                    if cstr in debug_disabled_categories:
                        disabled = True
                    if is_debug_enabled(cstr):
                        enabled = True
        self.debug_enabled = enabled and not disabled
        # ready, keep track of it:
        add_logger(self.categories, self)
        for x in categories:
            if x not in KNOWN_FILTERS:
                self.warn("unknown logging category: %s", x)
        if self.debug_enabled:
            self.debug(f"debug enabled for {self.categories}")

    def get_info(self) -> dict[str, Any]:
        return {
            "categories": self.categories,
            "debug": self.debug_enabled,
            "level": self._logger.getEffectiveLevel(),
        }

    def __repr__(self):
        return f"Logger{self.categories}"

    def getEffectiveLevel(self) -> int:
        return self._logger.getEffectiveLevel()

    def setLevel(self, level: int) -> None:
        self.level = level
        self._logger.setLevel(level)

    def is_debug_enabled(self) -> bool:
        return self.debug_enabled

    def enable_debug(self) -> None:
        self.debug_enabled = True

    def disable_debug(self) -> None:
        self.debug_enabled = False

    def critical(self, enable=False) -> None:
        self.level_override = logging.CRITICAL if enable else 0

    def log(self, level, msg: str, *args, **kwargs) -> None:
        if level <= self.min_level:
            return
        exc_info = kwargs.get("exc_info", None)
        if exc_info is True:
            kwargs.pop("exc_info")
            ei = sys.exc_info()
            if ei != (None, None, None):
                exc_info = False
                kwargs["exc_info"] = ei
        if LOG_PREFIX:
            msg = LOG_PREFIX + msg
        frame = kwargs.pop("frame", None)
        backtrace = kwargs.pop("backtrace", level >= BACKTRACE_LEVEL) or bool(frame)
        if backtrace or any(exp.match(msg) for exp in backtrace_expressions):
            import traceback
            tb = traceback.extract_stack(frame)
            count = len(tb)
            for i, frame in enumerate(tb):
                if frame.filename.endswith(MODULE_FILE) and i >= (count - 3):
                    # skip the logger's own calls
                    break
                try:
                    frame_summary = tb.format_frame_summary(frame)
                except AttributeError:
                    # `format_frame_summary` requires Python 3.11
                    pass
                else:
                    for rec in frame_summary.splitlines():
                        global_logging_handler(self._logger.log, self.level_override or level, rec)
        global_logging_handler(self._logger.log, self.level_override or level, msg, *args, **kwargs)

    def __call__(self, msg: str, *args, **kwargs) -> None:
        if self.debug_enabled:
            self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.INFO, msg, *args, **kwargs)

    def warn(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.WARN, msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        self.log(logging.ERROR, msg, *args, **kwargs)

    def estr(self, e, **kwargs) -> None:
        einfo = str(e) or type(e)
        self.error(f" {einfo}", **kwargs)

    def handle(self, record) -> None:
        self.log(record.levelno, record.msg, *record.args, exc_info=record.exc_info)

    def trap_error(self, message: str, *args) -> AbstractContextManager:
        return ErrorTrapper(self, message, args)


class ErrorTrapper(AbstractContextManager):
    def __init__(self, logger, message, args):
        self.logger = logger
        self.message = message
        self.args = args

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.logger.error(self.message, *self.args, exc_info=(exc_type, exc_val, exc_tb), backtrace=True)
            return True

    def __repr__(self):
        return "ErrorTrapper"


# we want to keep a reference to all the loggers in use,
# and we may have multiple loggers for the same key,
# but we don't want to prevent garbage collection so use a list of `weakref`s
all_loggers: dict[str, set['weakref.ReferenceType[Logger]']] = {}


def add_logger(categories, logger: Logger) -> None:
    categories = list(categories)
    categories.append("all")
    ref_logger = weakref.ref(logger)
    for cat in categories:
        all_loggers.setdefault(cat, set()).add(ref_logger)


def get_all_loggers() -> set[Logger]:
    a = set()
    for loggers_set in all_loggers.values():
        for logger in tuple(loggers_set):
            # weakref:
            instance = logger()
            if instance:
                a.add(instance)
    return a


def get_loggers_for_categories(*cat) -> list[Logger]:
    if not cat:
        return []
    if "all" in cat:
        return list(get_all_loggers())
    cset = set(cat)
    matches = set()
    for logger in get_all_loggers():
        if set(logger.categories).issuperset(cset):
            matches.add(logger)
    return list(matches)


def enable_debug_for(*cat) -> list[Logger]:
    loggers: list[Logger] = []
    for logger in get_loggers_for_categories(*cat):
        if not logger.is_debug_enabled():
            logger.enable_debug()
            loggers.append(logger)
    return loggers


def disable_debug_for(*cat) -> list[Logger]:
    loggers: list[Logger] = []
    for logger in get_loggers_for_categories(*cat):
        if logger.is_debug_enabled():
            logger.disable_debug()
            loggers.append(logger)
    return loggers


class CaptureHandler(logging.Handler):
    def __init__(self):
        super().__init__(logging.DEBUG)
        self.records = []

    def handle(self, record) -> None:
        self.records.append(record)

    def emit(self, record) -> None:
        self.records.append(record)

    def createLock(self) -> None:
        self.lock = None


class SIGPIPEStreamHandler(logging.StreamHandler):
    def flush(self) -> None:
        try:
            super().flush()
        except BrokenPipeError:
            pass

    def emit(self, record) -> None:
        # noinspection PyBroadException
        try:
            msg = self.format(record)
            stream = self.stream
            # issue 35046: merged two stream.writes into one.
            stream.write(msg + self.terminator)
            self.flush()
        except RecursionError:  # See issue 36272
            raise
        except BrokenPipeError:
            pass
        except Exception:
            self.handleError(record)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.x11.bindings.wait_for_x_server import wait_for_x_server       #@UnresolvedImport
import os

wait_for_x_server(os.environ.get("DISPLAY", ":0"), 5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import unittest

from xpra.util.str_fn import Ellipsizer, memoryview_to_bytes as mtb
from xpra.log import consume_verbose_argv, Logger

log = Logger("lz4")


def e(s):
    return Ellipsizer(mtb(s))


INVALID_INPUTS = (None, True, 1, 1.2, [1, 2], (1, 2), object())


class TestLZ4(unittest.TestCase):

    def test_libversions(self):
        from xpra.net.lz4.lz4 import get_version  # @UnresolvedImport
        assert get_version() >= (1, 8)

    def td(self, v, match_value=None, maxsize=512 * 1024):
        log("tc%s", (Ellipsizer(v), Ellipsizer(match_value), maxsize))
        from xpra.net.lz4.lz4 import decompress  # @UnresolvedImport
        value = decompress(v, maxsize)
        if match_value is not None:
            assert mtb(value) == mtb(match_value), "expected %s but got %s" % (e(match_value), e(value))
        return value

    def fd(self, v, match_value=None, maxsize=512 * 1024):
        try:
            self.td(v, match_value, maxsize)
        except Exception:
            pass
        else:
            raise ValueError("decompression should have failed for %r" % v)

    def tc(self, v, match_value=None, level=2, maxsize=512 * 1024):
        log("tc%s", (Ellipsizer(v), Ellipsizer(match_value), level, maxsize))
        from xpra.net.lz4.lz4 import compress  # @UnresolvedImport
        value = compress(v, 10 - level)
        if match_value is not None:
            assert mtb(value) == mtb(match_value), "expected %s but got %s" % (e(match_value), e(value))
        return value

    def fc(self, v, match_value=None, maxsize=512 * 1024):
        try:
            self.tc(v, match_value, maxsize)
        except Exception:
            pass
        else:
            raise ValueError("compression should have failed for %r" % v)

    def test_decompressinvalidinput(self):
        for v in INVALID_INPUTS:
            self.fd(v)
        self.fd(b"hello")

    def test_decompressinputtoosmall(self):
        self.fd(b"hello")

    def test_decompressvalid(self):
        self.td(b"\x05\x00\x00\x00Phello", b"hello")

    def test_limit(self):
        br = b"".join((
            b"\x00\x00\x10\x00\x1f\x00\x01\x00",
            (b"\xff" * 4111),
            b"\xf6P\x00\x00\x00\x00\x00"))
        self.td(br, b"\0" * 1024 * 1024, 1024 * 1024)
        self.fd(br, b"\0" * 1024 * 1024, 1024 * 1024 - 1)

    def test_compress(self):
        for l in range(2, 11):
            self.tc(b"hello", b"\x05\x00\x00\x00Phello", l)

    def test_compressinvalidinput(self):
        for v in INVALID_INPUTS:
            self.fc(v)

    def test_roundtrip(self):
        TEST_INPUT = [b"hello", b"*" * 1024, b"+" * 64 * 1024]

        #find some real "text" files:

        def addf(path):
            if path and os.path.exists(path):
                with open(path, "rb") as f:
                    TEST_INPUT.append(f.read())

        if __file__:
            addf(__file__)
            path = os.path.abspath(os.path.dirname(__file__))
            while path:
                addf(os.path.join(path, "COPYING"))
                addf(os.path.join(path, "README.md"))
                parent_path = os.path.abspath(os.path.join(path, os.pardir))
                if parent_path == path or not parent_path:
                    break
                path = parent_path
        for l in range(2, 11):
            for v in TEST_INPUT:
                c = self.tc(v, level=l)
                self.td(c, v)


def main() -> None:
    consume_verbose_argv(sys.argv, "lz4")
    try:
        from xpra.net.lz4.lz4 import decompress, compress
        assert decompress and compress
    except ImportError as e:
        print("lz4 test skipped: %s" % e)
    else:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from time import monotonic
from collections import deque
from collections.abc import Sequence

from xpra.codecs.image import ImageWrapper
from xpra.util.objects import typedict
from xpra.util.str_fn import memoryview_to_bytes
from xpra.log import Logger

log = Logger("encoder", "proxy")


def get_version() -> Sequence[int]:
    return 0, 2


def get_type() -> str:
    return "proxy"


def get_info() -> dict[str, Any]:
    return {"version": get_version()}


def get_encodings() -> Sequence[str]:
    return ("proxy",)


class Encoder:
    """
        This is a "fake" encoder which just forwards
        the raw pixels and the metadata that goes with it.
    """

    def init_context(self, encoding: str, width: int, height: int, src_format: str, options: typedict) -> None:
        self.encoding = encoding
        self.width = width
        self.height = height
        self.src_format = src_format
        self.dst_formats = options.strtupleget("dst-formats")
        self.last_frame_times: deque[float] = deque(maxlen=200)
        self.frames = 0
        self.time = 0
        self.first_frame_timestamp = 0

    def is_ready(self) -> bool:
        return True

    def get_info(self) -> dict[str, Any]:
        info = get_info()
        if not self.src_format:
            return info
        info.update({
            "frames": self.frames,
            "width": self.width,
            "height": self.height,
            "encoding": self.encoding,
            "src_format": self.src_format,
            "dst_formats": self.dst_formats,
        })
        # calculate fps:
        now = monotonic()
        last_time = now
        cut_off = now - 10.0
        f = 0
        for v in tuple(self.last_frame_times):
            if v > cut_off:
                f += 1
                last_time = min(last_time, v)
        if f > 0 and last_time < now:
            info["fps"] = int(0.5 + f / (now - last_time))
        return info

    def __repr__(self):
        if not self.src_format:
            return "proxy_encoder(uninitialized)"
        return f"proxy_encoder({self.src_format} - {self.width}x{self.height})"

    def is_closed(self) -> bool:
        return not bool(self.src_format)

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return "proxy"

    def get_src_format(self) -> str:
        return self.src_format

    def clean(self) -> None:
        self.width = 0
        self.height = 0
        self.src_format = ""
        self.encoding = ""
        self.src_format = ""
        self.dst_formats = []
        self.last_frame_times = deque()
        self.frames = 0
        self.time = 0
        self.first_frame_timestamp = 0

    def compress_image(self, image: ImageWrapper, options: typedict) -> tuple[bytes, dict]:
        log("compress_image(%s, %s)", image, options)
        # pass the pixels as they are
        if image.get_planes() != ImageWrapper.PACKED:
            raise RuntimeError(f"invalid number of planes: {image.get_planes()}")
        pixels = image.get_pixels()
        if not pixels:
            raise RuntimeError(f"failed to get pixels from {image}")
        # info used by proxy encoder:
        client_options = {
            "proxy": True,
            "frame": self.frames,
            "pts": image.get_timestamp() - self.first_frame_timestamp,
            "timestamp": image.get_timestamp(),
            "rowstride": image.get_rowstride(),
            "depth": image.get_depth(),
            "rgb_format": image.get_pixel_format(),
            # pass-through encoder options:
            "options": options or {},
        }
        if self.frames == 0:
            self.first_frame_timestamp = image.get_timestamp()
            # must pass dst_formats so the proxy can instantiate the video encoder
            # with the correct CSC config:
            client_options["dst_formats"] = self.dst_formats
        log("compress_image(%s, %s) returning %s bytes and options=%s", image, options, len(pixels), client_options)
        self.last_frame_times.append(monotonic())
        self.frames += 1
        return memoryview_to_bytes(pixels[:]), client_options



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import Any
from collections.abc import Callable
from cairo import Context, ImageSurface, Format, Operator

from xpra.client.gui.paint_colors import get_paint_box_color
from xpra.client.gui.window_backing_base import WindowBackingBase, fire_paint_callbacks
from xpra.common import roundup, PaintCallbacks
from xpra.util.str_fn import memoryview_to_bytes
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.log import Logger

GLib = gi_import("GLib")
Gdk = gi_import("Gdk")

log = Logger("paint", "cairo")

COPY_OLD_BACKING = envbool("XPRA_CAIRO_COPY_OLD_BACKING", True)

FORMATS = {-1: "INVALID"}
for attr in dir(Format):
    if attr.isupper():
        FORMATS[getattr(Format, attr)] = attr


def cairo_paint_pointer_overlay(context, cursor_data, px: int, py: int, start_time) -> None:
    if not cursor_data:
        return
    elapsed = max(0, monotonic() - start_time)
    if elapsed > 6:
        return
    # pylint: disable=import-outside-toplevel
    try:
        from xpra.client.gtk3.cairo_image import make_image_surface
    except ImportError:
        return

    cw = cursor_data[3]
    ch = cursor_data[4]
    xhot = cursor_data[5]
    yhot = cursor_data[6]
    pixels = cursor_data[8]
    x = px - xhot
    y = py - yhot

    alpha = max(0.0, (5.0 - elapsed) / 5.0)
    log("cairo_paint_pointer_overlay%s drawing pointer with cairo, alpha=%s",
        (context, x, y, start_time), alpha)
    bgra = memoryview_to_bytes(pixels)
    img = make_image_surface(Format.ARGB32, "BGRA", bgra, cw, ch, cw * 4)
    context.translate(x, y)
    context.set_source_surface(img, 0, 0)
    context.set_operator(Operator.OVER)
    context.paint_with_alpha(alpha)


class CairoBackingBase(WindowBackingBase):
    HAS_ALPHA = envbool("XPRA_ALPHA", True)

    def __init__(self, wid: int, window_alpha: bool, _pixel_depth=0):
        super().__init__(wid, window_alpha and self.HAS_ALPHA)
        self.size = 0, 0
        self.render_size = 0, 0
        self.fps_image = None

    def init(self, ww: int, wh: int, bw: int, bh: int) -> None:
        mod = self.size != (bw, bh) or self.render_size != (ww, wh)
        self.size = bw, bh
        self.render_size = ww, wh
        if mod:
            self.create_surface()

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        info |= {
            "type": "Cairo",
            "rgb-formats": self.get_rgb_formats(),
        }
        return info

    def create_surface(self) -> Context | None:
        bw, bh = self.size
        old_backing = self._backing
        # should we honour self.depth here?
        self._backing = None
        if bw == 0 or bh == 0:
            # this can happen during cleanup
            return None
        backing = ImageSurface(Format.ARGB32, bw, bh)
        self._backing = backing
        cr = Context(backing)
        if self._alpha_enabled:
            cr.set_operator(Operator.CLEAR)
            cr.set_source_rgba(1, 1, 1, 0)
        else:
            cr.set_operator(Operator.SOURCE)
            cr.set_source_rgba(1, 1, 1, 1)
        cr.rectangle(0, 0, bw, bh)
        cr.fill()
        if COPY_OLD_BACKING and old_backing is not None:
            cr.save()
            oldw, oldh = old_backing.get_width(), old_backing.get_height()
            sx, sy, dx, dy, w, h = self.gravity_copy_coords(oldw, oldh, bw, bh)
            cr.translate(dx - sx, dy - sy)
            cr.rectangle(sx, sy, w, h)
            cr.clip()
            cr.set_operator(Operator.SOURCE)
            cr.set_source_surface(old_backing, 0, 0)
            cr.paint()
            cr.restore()
            if self.paint_box_line_width > 0 and (oldw != bw or oldh != bh):
                if bw > oldw:
                    self.cairo_paint_box(cr, "padding", oldw, 0, bw - oldw, bh)
                if bh > oldh:
                    self.cairo_paint_box(cr, "padding", 0, oldh, bw, bh - oldh)
            backing.flush()
        return cr

    def close(self) -> None:
        backing = self._backing
        if backing:
            backing.finish()
            self._backing = None
        super().close()

    def cairo_paint_pixbuf(self, pixbuf, x: int, y: int, options) -> None:
        """ must be called from UI thread """
        log("source pixbuf: %s", pixbuf)
        w, h = pixbuf.get_width(), pixbuf.get_height()
        self.cairo_paint_from_source(Gdk.cairo_set_source_pixbuf, pixbuf, x, y, w, h, w, h, options)

    def cairo_paint_surface(self, img_surface, x: int, y: int, width: int, height: int, options) -> None:
        iw, ih = img_surface.get_width(), img_surface.get_height()
        log("source image surface: %s",
            (img_surface.get_format(), iw, ih, img_surface.get_stride(), img_surface.get_content(),))

        def set_source_surface(gc, surface, sx: int, sy: int) -> None:
            gc.set_source_surface(surface, sx, sy)

        self.cairo_paint_from_source(set_source_surface, img_surface, x, y, iw, ih, width, height, options)

    def cairo_paint_from_source(self, set_source_fn: Callable[[Any, Any, int, int], None], source,
                                x: int, y: int, iw: int, ih: int, width: int, height: int, options) -> None:
        """ must be called from UI thread """
        backing = self._backing
        log("cairo_paint_surface%s backing=%s, paint box line width=%i",
            (set_source_fn, source, x, y, iw, ih, width, height, options),
            backing, self.paint_box_line_width)
        if not backing:
            return
        gc = Context(backing)
        if self.paint_box_line_width:
            gc.save()

        gc.rectangle(x, y, width, height)
        gc.clip()

        gc.set_operator(Operator.CLEAR)
        gc.rectangle(x, y, width, height)
        gc.fill()

        gc.set_operator(Operator.SOURCE)
        gc.translate(x, y)
        if iw != width or ih != height:
            gc.scale(width / iw, height / ih)
        gc.rectangle(0, 0, width, height)
        set_source_fn(gc, source, 0, 0)
        gc.paint()

        if self.paint_box_line_width:
            gc.restore()
            encoding = options.get("encoding")
            self.cairo_paint_box(gc, encoding, x, y, width, height)

        flush = options.get("flush", 0)
        if flush == 0:
            self.record_fps_event()

    def cairo_paint_box(self, gc, encoding: str, x: int, y: int, w: int, h: int) -> None:
        color = get_paint_box_color(encoding)
        gc.set_line_width(self.paint_box_line_width)
        gc.set_source_rgba(*color)
        gc.rectangle(x, y, w, h)
        gc.stroke()

    def do_paint_rgb(self, context, encoding: str, rgb_format: str, img_data,
                     x: int, y: int, width: int, height: int, render_width: int, render_height: int, rowstride: int,
                     options: typedict, callbacks: PaintCallbacks) -> None:
        """ must be called from the UI thread
            this method is only here to ensure that we always fire the callbacks,
            the actual paint code is in _do_paint_rgb[16|24|30|32]
        """
        if not options.boolget("paint", True):
            fire_paint_callbacks(callbacks)
            return
        if self._backing is None:
            fire_paint_callbacks(callbacks, -1, "no backing")
            return
        x, y = self.gravity_adjust(x, y, options)
        if rgb_format == "r210":
            bpp = 30
        elif rgb_format == "BGR565":
            bpp = 16
        else:
            bpp = len(rgb_format) * 8  # ie: "BGRA" -> 32
        if rowstride == 0:
            rowstride = width * roundup(bpp, 8) // 8
        try:
            fmt = {
                16: Format.RGB16_565,
                24: Format.RGB24,
                30: Format.RGB30,
                32: Format.ARGB32 if self._alpha_enabled else Format.RGB24,
            }.get(bpp, Format.INVALID)
            if fmt == Format.INVALID:
                raise ValueError(f"invalid rgb format {rgb_format!r} with bit depth {bpp}")
            options["rgb_format"] = rgb_format
            alpha = bpp == 32 and self._alpha_enabled
            self._do_paint_rgb(fmt, alpha, img_data,
                               x, y, width, height, render_width, render_height, rowstride, options)
            fire_paint_callbacks(callbacks, True)
        except Exception as e:
            if not self._backing:
                fire_paint_callbacks(callbacks, -1, "paint error on closed backing ignored")
            else:
                log.error("Error painting rgb%s", bpp, exc_info=True)
                message = f"paint rgb{bpp} error: {e}"
                fire_paint_callbacks(callbacks, False, message)

    def _do_paint_rgb(self, *args) -> None:
        # see CairoBacking
        raise NotImplementedError()

    def paint_scroll(self, img_data, options: typedict, callbacks: PaintCallbacks) -> None:
        # newer servers use an option,
        # older ones overload the img_data:
        scroll_data = options.tupleget("scroll", img_data)
        GLib.idle_add(self.do_paint_scroll, scroll_data, callbacks)

    def do_paint_scroll(self, scrolls, callbacks: PaintCallbacks) -> None:
        old_backing = self._backing
        if not old_backing:
            fire_paint_callbacks(callbacks, False, message="no backing")
            return
        gc = self.create_surface()
        if not gc:
            fire_paint_callbacks(callbacks, False, message="no context")
            return
        gc.set_operator(Operator.SOURCE)
        for sx, sy, sw, sh, xdelta, ydelta in scrolls:
            gc.set_source_surface(old_backing, xdelta, ydelta)
            x = sx + xdelta
            y = sy + ydelta
            gc.rectangle(x, y, sw, sh)
            gc.fill()
            if self.paint_box_line_width > 0:
                self.cairo_paint_box(gc, "scroll", x, y, sw, sh)
        del gc
        self._backing.flush()
        fire_paint_callbacks(callbacks)

    def cairo_draw(self, context) -> None:
        backing = self._backing
        log("cairo_draw: backing=%s, size=%s, render-size=%s, offsets=%s, pointer_overlay=%s",
            backing, self.size, self.render_size, self.offsets, self.pointer_overlay)
        if backing is None:
            return
        # try:
        #    log("clip rectangles=%s", context.copy_clip_rectangle_list())
        # except:
        #    log.error("clip:", exc_info=True)
        ww, wh = self.render_size
        w, h = self.size
        if ww == 0 or w == 0 or wh == 0 or h == 0:
            return
        if w != ww or h != wh:
            context.scale(ww / w, wh / h)
        x, y = self.offsets[:2]
        if x != 0 or y != 0:
            context.translate(x, y)
        context.set_operator(Operator.SOURCE)
        context.set_source_surface(backing, 0, 0)
        context.paint()

        if self.pointer_overlay and self.cursor_data:
            px, py, _size, start_time = self.pointer_overlay[2:]
            spx = round(w * px / ww)
            spy = round(h * py / wh)
            cairo_paint_pointer_overlay(context, self.cursor_data, x + spx, y + spy, start_time)

        if self.is_show_fps() and self.fps_image:
            x, y = 10, 10
            context.translate(x, y)
            context.set_operator(Operator.OVER)
            context.set_source_surface(self.fps_image, 0, 0)
            context.paint()
            self.cancel_fps_refresh()

            def refresh_screen() -> None:
                self.fps_refresh_timer = 0
                b = self._backing
                if b:
                    self.update_fps()
                    rw, rh = self.render_size
                    self.repaint(0, 0, rw, rh)

            self.cancel_fps_refresh()
            self.fps_refresh_timer = GLib.timeout_add(1000, refresh_screen)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import secrets
from struct import pack
from typing import Any
from collections.abc import Iterable, Sequence

from xpra.util.str_fn import csv, print_nested_dict, strtobytes, hexstr
from xpra.util.env import envint, envbool
from xpra.util.version import parse_version
from xpra.net.digest import get_salt
from xpra.log import Logger, consume_verbose_argv

log = Logger("network", "crypto")

cryptography = None

ENCRYPT_FIRST_PACKET = envbool("XPRA_ENCRYPT_FIRST_PACKET", False)

DEFAULT_IV = os.environ.get("XPRA_CRYPTO_DEFAULT_IV", "0000000000000000")
DEFAULT_SALT = os.environ.get("XPRA_CRYPTO_DEFAULT_SALT", "0000000000000000").encode("latin1")
DEFAULT_ITERATIONS = envint("XPRA_CRYPTO_DEFAULT_ITERATIONS", 10000)
DEFAULT_KEYSIZE = envint("XPRA_CRYPTO_KEYSIZE", 32)
if DEFAULT_KEYSIZE not in (16, 24, 32):
    log.warn("Warning: default key size %i (%i bits) is not supported",
             DEFAULT_KEYSIZE, DEFAULT_KEYSIZE * 8)
# these were made configurable in xpra 4.3:
MIN_ITERATIONS = envint("XPRA_CRYPTO_STRETCH_MIN_ITERATIONS", 1000)
MAX_ITERATIONS = envint("XPRA_CRYPTO_STRETCH_MIN_ITERATIONS", 1000000)
DEFAULT_MODE = os.environ.get("XPRA_CRYPTO_MODE", "CBC")
DEFAULT_KEY_HASH = os.environ.get("XPRA_CRYPTO_KEY_HASH", "SHA1")
DEFAULT_ALWAYS_PAD = envbool("XPRA_CRYPTO_ALWAYS_PAD", False)
DEFAULT_STREAM = envbool("XPRA_CRYPTO_STREAM", True)
DEFAULT_KEY_STRETCH = "PBKDF2"

PADDING_PKCS7 = "PKCS#7"
ALL_PADDING_OPTIONS = (PADDING_PKCS7, )
INITIAL_PADDING = os.environ.get("XPRA_CRYPTO_INITIAL_PADDING", PADDING_PKCS7)
DEFAULT_PADDING = PADDING_PKCS7
PREFERRED_PADDING = os.environ.get("XPRA_CRYPTO_PREFERRED_PADDING", PADDING_PKCS7)
if PREFERRED_PADDING not in ALL_PADDING_OPTIONS:
    raise ValueError(f"invalid preferred padding: {PREFERRED_PADDING}")
if INITIAL_PADDING not in ALL_PADDING_OPTIONS:
    raise ValueError(f"invalid padding: {INITIAL_PADDING}")


# make sure the preferred one is first in the list:


def get_padding_options() -> Sequence[str]:
    options = [PREFERRED_PADDING]
    for x in ALL_PADDING_OPTIONS:
        if x not in options:
            options.append(x)
    return tuple(options)


PADDING_OPTIONS: Sequence[str] = get_padding_options()

# pylint: disable=import-outside-toplevel
CIPHERS: Sequence[str] = ()
MODES: Sequence[str] = ()
KEY_HASHES: Sequence[str] = ()
KEY_STRETCHING: Sequence[str] = ()


def crypto_backend_init():
    global cryptography, CIPHERS, MODES, KEY_HASHES, KEY_STRETCHING
    if cryptography:
        log("cryptography %s found in %s", cryptography.__version__, cryptography)
        return cryptography
    try:
        import cryptography as pc
        cryptography = pc
        MODES = tuple(x for x in os.environ.get(
            "XPRA_CRYPTO_MODES",
            "CBC,CFB,CTR").split(",") if x in ("CBC", "CFB", "CTR"))
        KEY_HASHES = ("SHA1", "SHA224", "SHA256", "SHA384", "SHA512")
        KEY_STRETCHING = ("PBKDF2",)
        CIPHERS = ("AES",)
        from cryptography.hazmat.backends import default_backend
        backend = default_backend()
        log("default_backend()=%s", backend)
        log("backends=%s", getattr(backend, "_backends", []))
        from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
        from cryptography.hazmat.primitives import hashes
        assert Cipher and algorithms and modes and hashes  # type: ignore[truthy-function]
        validate_backend()
        name = getattr(backend, "name", "")
        if name:
            log.info(f"cryptography {cryptography.__version__} using {name!r} backend")
        return cryptography
    except ImportError:
        log("crypto backend init failure", exc_info=True)
        log.error("Error: cannot import python-cryptography")
    except Exception:
        log.error("Error: cannot initialize python-cryptography", exc_info=True)
    cryptography = None
    CIPHERS = MODES = KEY_HASHES = KEY_STRETCHING = ()
    return None


def get_ciphers() -> Sequence[str]:
    return CIPHERS


def get_modes() -> Sequence[str]:
    return MODES


def get_key_hashes() -> Sequence[str]:
    return KEY_HASHES


def validate_backend() -> None:
    log("validate_backend() will validate AES modes: " + csv(MODES))
    message = b"some message1234" * 8
    key_data = b"this is our secret"
    key_salt = DEFAULT_SALT
    iterations = DEFAULT_ITERATIONS
    for mode in MODES:
        log("testing AES-%s", mode)
        key = None
        for key_hash in KEY_HASHES:
            key = get_key(key_data, key_salt, key_hash, DEFAULT_KEYSIZE, iterations)
            assert key
        block_size = get_block_size(mode)
        log(" key=%s, block_size=%s", hexstr(key), block_size)
        assert key is not None, "pycryptography failed to generate a key"
        iv = strtobytes(DEFAULT_IV)
        enc_cipher = get_cipher(key, iv, mode)
        log(" encryptor cipher=%s", enc_cipher)
        assert enc_cipher is not None, "pycryptography failed to generate an encryptor"
        dec_cipher = get_cipher(key, iv, mode)
        log(" decryptor cipher=%s", dec_cipher)
        assert dec_cipher is not None, "pycryptography failed to generate a decryptor"
        test_messages = [message * (1 + block_size)]
        if block_size == 0:
            test_messages.append(message[:29])
        else:
            test_messages.append(message[:block_size])
        try:
            for m in test_messages:
                enc = enc_cipher.encryptor()
                ev = enc.update(m) + enc.finalize()
                evs = hexstr(ev)
                log(" encrypted(%s)=%s", m, evs)
                dec = dec_cipher.decryptor()
                dv = dec.update(ev) + dec.finalize()
                log(" decrypted(%s)=%s", evs, dv)
                if dv != m:
                    raise RuntimeError(f"expected {m!r} but got {dv!r}")
                log(" test passed")
        except ValueError:
            print(f"test error on {mode=}")
            raise


def pad(padding: str, size: int) -> bytes:
    if padding != PADDING_PKCS7:
        raise ValueError(f"invalid padding: {padding!r}")
    return pack("B", size) * size


def choose_padding(options: Iterable[str]) -> str:
    if PREFERRED_PADDING in options:
        return PREFERRED_PADDING
    for x in options:
        if x in PADDING_OPTIONS:
            return x
    raise ValueError(f"cannot find a valid padding in {options}")


def get_iv() -> str:
    return secrets.token_urlsafe(16)[:16]


def get_iterations() -> int:
    return DEFAULT_ITERATIONS


def new_cipher_caps(proto, cipher: str, cipher_mode: str, encryption_key,
                    padding_options, always_pad: bool, stream: bool) -> dict[str, Any]:
    iv = get_iv()
    key_salt = get_salt()
    key_size = DEFAULT_KEYSIZE
    key_hash = DEFAULT_KEY_HASH
    key_stretch = DEFAULT_KEY_STRETCH
    iterations = get_iterations()
    padding = choose_padding(padding_options)
    proto.set_cipher_in(cipher + "-" + cipher_mode, strtobytes(iv),
                        encryption_key, key_salt, key_hash, key_size,
                        iterations, padding, always_pad, stream)
    return {
        "cipher": cipher,
        "mode": cipher_mode,
        "mode.options": MODES,
        "iv": iv,
        "key_salt": key_salt,
        "key_hash": key_hash,
        "key_size": key_size,
        "key_stretch": key_stretch,
        "key_stretch.options": KEY_STRETCHING,
        "key_stretch_iterations": iterations,
        "stream": stream,
        "padding": padding,
        "padding.options": PADDING_OPTIONS,
    }


def get_crypto_caps(full=True) -> dict[str, Any]:
    crypto_backend_init()
    caps: dict[str, Any] = {
        "padding": {"options": PADDING_OPTIONS},
        "modes": {"options": MODES},
        "stretch": {"options": KEY_STRETCHING},
    }
    if full and cryptography:
        caps["python-cryptography"] = {
            "": True,
            "version": parse_version(cryptography.__version__),
        }
    return caps


def get_mode(ciphername: str) -> str:
    mode = (ciphername + "-").split("-")[1] or DEFAULT_MODE
    if not ciphername.startswith("AES"):
        raise ValueError(f"unsupported cipher {ciphername!r}")
    if mode not in MODES:
        raise ValueError(f"unsupported AES mode {mode!r}")
    return mode


def get_cipher(key: bytes, iv: bytes, mode: str = DEFAULT_MODE):
    if not iv:
        raise ValueError("missing encryption iv")
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
    mode_class = getattr(modes, mode, None)
    if mode_class is None:
        raise ValueError(f"no {mode} mode in this version of python-cryptography")
    from cryptography.hazmat.backends import default_backend
    return Cipher(algorithms.AES(key), mode_class(iv), backend=default_backend())


def get_block_size(mode: str) -> int:
    if mode == "CBC":
        # older versions require 32
        return 16
    return 0


def get_key(key_data: bytes, key_salt: bytes, key_hash: str, key_size: int, iterations: int) -> bytes:
    assert key_size >= 16
    if iterations < MIN_ITERATIONS or iterations > MAX_ITERATIONS:
        raise ValueError(f"invalid number of iterations {iterations}, range is {MIN_ITERATIONS} to {MAX_ITERATIONS}")
    if not key_data:
        raise ValueError("missing encryption key data")
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from cryptography.hazmat.primitives import hashes
    if key_hash.upper() not in KEY_HASHES:
        raise ValueError(f"invalid key hash {key_hash.upper()!r}, should be one of: " + csv(KEY_HASHES))
    algorithm = getattr(hashes, key_hash.upper(), None)
    if not algorithm:
        raise ValueError(f"{key_hash.upper()!r} not found in cryptography hashes")
    hash_algo = algorithm()
    from cryptography.hazmat.backends import default_backend
    kdf = PBKDF2HMAC(algorithm=hash_algo, length=key_size,
                     salt=strtobytes(key_salt), iterations=iterations,
                     backend=default_backend())
    key = kdf.derive(key_data)
    return key


def main():
    from xpra.platform import program_context
    with program_context("Encryption Properties"):
        consume_verbose_argv(sys.argv, "crypto")
        crypto_backend_init()
        print_nested_dict(get_crypto_caps())


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.util.str_fn import csv


class KeyEvent:
    __slots__ = ("modifiers", "keyname", "keyval", "keycode", "group", "string", "pressed")

    def __init__(self):
        self.modifiers: list[str] = []
        self.keyname: str = ""
        self.keyval: int = 0
        self.keycode: int = 0
        self.group: int = 0
        self.string: str = ""
        self.pressed: bool = True

    def __repr__(self):
        strattrs = csv(f"{k}="+str(getattr(self, k)) for k in KeyEvent.__slots__)
        return f"KeyEvent({strattrs})"



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import socket
from typing import Any, Dict

from libc.stdint cimport uint32_t, uint16_t, uint8_t   # pylint: disable=syntax-error

from xpra.os_util import load_binary_file, LINUX
from xpra.log import Logger

log = Logger("util", "network")

ctypedef uint32_t __u32
ctypedef uint16_t __u16
ctypedef uint8_t __u8

DEF ETHTOOL_GDRVINFO = 0x00000003
DEF ETHTOOL_BUSINFO_LEN = 32


cdef extern from "linux/ethtool.h":
    int ETHTOOL_GSET
    int ETHTOOL_BUSINFO_LEN
    cdef struct ethtool_cmd:
        __u32   cmd
        __u32   supported
        __u32   advertising
        __u16   speed
        __u8    duplex
        __u8    port
        __u8    phy_address
        __u8    transceiver
        __u8    autoneg
        __u8    mdio_support
        __u32   maxtxpkt
        __u32   maxrxpkt
        __u16   speed_hi
        __u8    eth_tp_mdix
        __u8    eth_tp_mdix_ctrl
        __u32   lp_advertising
        __u32   reserved[2]

    cdef struct ethtool_drvinfo:
        uint32_t    cmd
        char    driver[32]     # driver short name, "tulip", "eepro100"
        char    version[32]     # driver version string
        char    fw_version[32]     # firmware version string, if applicable
        char    bus_info[ETHTOOL_BUSINFO_LEN]     # Bus info for this IF.
                    # For PCI devices, use pci_dev->slot_name.
        char    reserved1[32]
        char    reserved2[16]
        uint32_t    n_stats     # number of u64's from ETHTOOL_GSTATS
        uint32_t    testinfo_len
        uint32_t    eedump_len     # Size of data from ETHTOOL_GEEPROM (bytes)
        uint32_t    regdump_len     # Size of data from ETHTOOL_GREGS (bytes)


cdef extern from "linux/sockios.h":
    int SIOCETHTOOL

cdef extern from "net/if.h":
    DEF IFNAMSIZ=16
    cdef struct ifr_ifrn:
        char ifrn_name[IFNAMSIZ]
    cdef struct ifr_ifru:
        int ifru_flags
        int ifru_ivalue
        int ifru_mtu
        void *ifru_data
    cdef struct ifreq:
        ifr_ifrn ifr_ifrn
        ifr_ifru ifr_ifru

cdef extern from "sys/ioctl.h":
    int ioctl(int fd, unsigned long request, ...)

#linux kernel: if_arp.h
ARPHRD : Dict[int,str] = {
    0   : "netrom",
    1   : "ether",
    2   : "eether",
    3   : "ax25",
    4   : "pronet",
    5   : "chaos",
    6   : "ieee802",
    7   : "arcnet",
    8   : "appletlk",
    15  : "dlci",
    19  : "atm",
    23  : "metricom",
    24  : "ieee1394",
    27  : "eui64",
    32  : "infiniband",

    256 : "slip",
    257 : "cslip",
    258 : "slip6",
    259 : "cslip6",
    260 : "rsrvd",
    264 : "adapt",
    270 : "rose",
    271 : "x25",
    272 : "hwx25",
    280 : "can",
    512 : "ppp",
    513 : "cisco",
    516 : "lapb",
    517 : "ddcmp",
    518 : "rawhdlc",

    768 : "tunnel",
    769 : "tunner6",
    770 : "frad",
    771 : "skip",
    772 : "loopback",
    773 : "localtlk",
    774 : "fddi",
    775 : "bif",
    776 : "sit",
    777 : "ipddp",
    778 : "ipgre",
    779 : "pimreg",
    780 : "hippi",
    781 : "ash",
    782 : "econet",
    783 : "irda",

    784 : "fcpp",
    785 : "fcal",
    786 : "fcpl",
    787 : "fcfabric",
    800 : "ieee802_tr",
    801 : "ieee80211",
    802 : "ieee80211_prism",
    803 : "ieee80211_radiotap",
    804 : "ieee802154",

    820 : "phonet",
    821 : "phonet_pipe",
    822 : "caif",
}


cdef str s(const char *v):
    pytmp = v[:]
    try:
        return pytmp.decode()
    except:
        return str(v[:])


def get_interface_info(int sockfd, ifname) -> Dict[str,Any]:
    if sockfd==0:
        return {}
    info = {}
    adapter_type = None
    sysnetfs = f"/sys/class/net/{ifname}"
    if os.path.exists(sysnetfs) and os.path.isdir(sysnetfs):
        type_file = os.path.join(sysnetfs, "type")
        if os.path.exists(type_file):
            dev_type = s(load_binary_file(type_file).rstrip(b"\n\r"))
            if dev_type:
                try:
                    idev_type = int(dev_type)
                    adapter_type = ARPHRD.get(idev_type)
                except ValueError:
                    pass
    if not adapter_type:
        if ifname.startswith("lo"):
            adapter_type = "loopback"
        elif ifname.startswith("wl"):
            adapter_type = "wireless"
        elif ifname.startswith("eth") or ifname.startswith("en"):
            adapter_type = "ethernet"
        elif ifname.startswith("ww"):
            adapter_type = "wan"
        else:
            wireless_path = f"{sysnetfs}/wireless"
            if os.path.exists(wireless_path):
                adapter_type = "wireless"
    if adapter_type:
        info["adapter-type"] = adapter_type
    if sockfd>0:
        info.update(get_ethtool_info(sockfd, ifname))
    return info

def get_ethtool_info(int sockfd, ifname:str) -> Dict[str,Any]:
    if len(ifname)>=IFNAMSIZ:
        log.warn("Warning: invalid interface name '%s'", ifname)
        log.warn(" maximum length is %i", IFNAMSIZ)
        return {}
    cdef ifreq ifr
    cdef ethtool_cmd edata
    bifname = ifname.encode("latin1")
    cdef char *cifname = bifname
    ifr.ifr_ifrn.ifrn_name = cifname
    ifr.ifr_ifru.ifru_data = <void*> &edata
    edata.cmd = ETHTOOL_GSET
    cdef int r = ioctl(sockfd, SIOCETHTOOL, &ifr)
    info = {}
    if r >= 0:
        info["speed"] = edata.speed*1000*1000
        #info["duplex"] = duplex: DUPLEX_HALF, DUPLEX_FULL DUPLEX_NONE?
    else:
        log("no ethtool interface speed available for %s", ifname)
        return info
    cdef ethtool_drvinfo drvinfo
    drvinfo.cmd = ETHTOOL_GDRVINFO
    ifr.ifr_ifru.ifru_data = <void *> &drvinfo
    r = ioctl(sockfd, SIOCETHTOOL, &ifr)
    if r>=0:
        info["driver"] = s(drvinfo.driver)
        info["version"] = s(drvinfo.version)
        info["firmware-version"] = s(drvinfo.fw_version)
        info["bus-info"] = s(drvinfo.bus_info)
    else:
        log.info("no driver information for %s", ifname)
    return info


def get_socket_tcp_info(sock) -> Dict[str,Any]:
    if not LINUX:
        #should be added for BSDs
        return {}
    from ctypes import c_uint8, c_uint32, c_uint64
    ALL_FIELDS = (
        ("state",           c_uint8),
        ("ca_state",        c_uint8),
        ("retransmits",     c_uint8),
        ("probes",          c_uint8),
        ("backoff",         c_uint8),
        ("options",         c_uint8),
        ("snd_wscale",      c_uint8, 4),
        ("rcv_wscale",      c_uint8, 4),
        ("rto",             c_uint32),
        ("ato",             c_uint32),
        ("snd_mss",         c_uint32),
        ("rcv_mss",         c_uint32),
        ("unacked",         c_uint32),
        ("sacked",          c_uint32),
        ("lost",            c_uint32),
        ("retrans",         c_uint32),
        ("fackets",         c_uint32),
        ("last_data_sent",  c_uint32),
        ("last_ack_sent",   c_uint32),
        ("last_data_recv",  c_uint32),
        ("last_ack_recv",   c_uint32),
        ("pmtu",            c_uint32),
        ("rcv_ssthresh",    c_uint32),
        ("rtt",             c_uint32),
        ("rttvar",          c_uint32),
        ("snd_ssthresh",    c_uint32),
        ("snd_cwnd",        c_uint32),
        ("advmss",          c_uint32),
        ("reordering",      c_uint32),
        ("rcv_rtt",         c_uint32),
        ("rcv_space",       c_uint32),
        ("total_retrans",   c_uint32),
        ("pacing_rate",     c_uint64),
        ("max_pacing_rate", c_uint64),
        ("bytes_acked",     c_uint64),
        ("bytes_received",  c_uint64),
        ("segs_out",        c_uint32),
        ("segs_in",         c_uint32),
        ("notsent_bytes",   c_uint32),
        ("min_rtt",         c_uint32),
        ("data_segs_in",    c_uint32),
        ("data_segs_out",   c_uint32),
        ("delivery_rate",   c_uint64),
        ("busy_time",       c_uint64),
        ("rwnd_limited",    c_uint64),
        ("sndbuf_limited",  c_uint64),
        ("delivered",       c_uint32),
        ("delivered_ce",    c_uint32),
        ("bytes_sent",      c_uint64),
        ("bytes_retrans",   c_uint64),
        ("dsack_dups",      c_uint32),
        ("reord_seen",      c_uint32),
        ("rcv_ooopack",     c_uint32),
        ("snd_wnd",         c_uint32),
        )
    from xpra.net.socket_util import get_sockopt_tcp_info
    return get_sockopt_tcp_info(sock, socket.TCP_INFO, ALL_FIELDS)

def get_send_buffer_info(sock) -> Dict[str,Any]:
    import fcntl
    import struct
    send_buffer_size = sock.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
    info = {
        "sndbuf_size"   : send_buffer_size,
        }
    fno = sock.fileno()
    if fno>0:
        def getioctl(k):
            return struct.unpack("I", fcntl.ioctl(fno, k, b'\0\0\0\0'))[0]
        #SIOCINQ = 0x541B
        SIOCOUTQ = 0x5411
        info["sndbuf_unacked"] = getioctl(SIOCOUTQ)
        SIOCOUTQNSD = 0x894B
        info["sndbuf_bytes"] = getioctl(SIOCOUTQNSD)
    return info

def get_tcp_info(sock) -> Dict[str,Any]:
    info = {}
    info.update(get_socket_tcp_info(sock))
    info.update(get_send_buffer_info(sock))
    log("get_tcp_info(%s)=%s", sock, info)
    return info



#!/usr/bin/env python3

import gi
gi.require_version("Gtk", "3.0")  # @UndefinedVariable
gi.require_version("Gdk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk, Gdk, GLib  # @UnresolvedImport

from xpra.gtk.cursors import cursor_types

width = 400
height = 200
def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_size_request(width, height)
	window.connect("delete_event", Gtk.main_quit)

	names = list(cursor_types.keys())
	names = names[:2]+names[:2]+names[:2]+names[:2]+names[:2]+names[:2]+names[:2]+names[:2]+names[:2]
	def set_new_cursor():
		name = names.pop()
		cursor = Gdk.Cursor(cursor_types[name])
		print("setting cursor to %s: %s" % (name, cursor))
		window.get_window().set_cursor(cursor)
		if len(names)==0:
			Gtk.main_quit()
		return True

	window.show_all()
	GLib.timeout_add(5, set_new_cursor)
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.string cimport memset

from xpra.log import Logger

log = Logger("util", "exec")

DEF PROC_FILLSTATUS = 0x0020    #read status
DEF PROC_PID = 0x1000           #process id numbers ( 0   terminated )


cdef extern from "proc/readproc.h":
    ctypedef int pid_t
    ctypedef struct proc_t:
        int ppid
        int pgrp        #process group id
        int session     #session id
        int nlwp        #stat,status     number of threads, or 0 if no clue
        int tgid        #(special)       thread group ID, the POSIX PID (see also: tid)
        int tty         #stat            full device number of controlling terminal

    proc_t * get_proc_stats(pid_t pid, proc_t *p)

    ctypedef struct PROCTAB:
        pass

    PROCTAB* openproc(int flags, pid_t *pid)
    proc_t* readproc(PROCTAB *PT, proc_t *p)
    void closeproc(PROCTAB* PT)

def get_parent_pid(int pid) -> int:
    cdef proc_t proc_info
    memset(&proc_info, 0, sizeof(proc_t))
    cdef PROCTAB *pt_ptr = openproc(PROC_FILLSTATUS | PROC_PID, &pid)
    try:
        if readproc(pt_ptr, &proc_info):
            return proc_info.ppid
    finally:
        closeproc(pt_ptr);
    return 0



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import signal

from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import scaled_image, label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.signals import register_os_signals
from xpra.util.objects import typedict
from xpra.util.config import parse_user_config_file, update_config_attribute
from xpra.exit_codes import ExitValue
from xpra.os_util import gi_import
from xpra.log import Logger, enable_debug_for

Gtk = gi_import("Gtk")

log = Logger("exec")

START_NEW_COMMAND_CONFIG = "start_new_command.conf"


def update_config(prop: str, value: str) -> None:
    update_config_attribute(prop, value, dirname="tools", filename=START_NEW_COMMAND_CONFIG)


def load_config() -> dict:
    return parse_user_config_file("tools", START_NEW_COMMAND_CONFIG)


_instance = None


def getStartNewCommand(run_callback, can_share=False, xdg_menu=None):
    global _instance
    if _instance is None:
        _instance = StartNewCommand(run_callback, can_share, xdg_menu)
    return _instance


def btn(label, tooltip, callback, icon_name="") -> Gtk.Button:
    b = Gtk.Button(label=label)
    b.set_tooltip_text(tooltip)
    b.connect("clicked", callback)
    icon = get_icon_pixbuf(icon_name)
    if icon:
        b.set_image(scaled_image(icon, 24))
    return b


class StartNewCommand:

    def __init__(self, run_callback=None, can_share=False, xdg_menu=None):
        self.run_callback = run_callback
        self.xdg_menu = typedict(xdg_menu or {})
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.close)
        self.window.set_default_size(400, 150)
        self.window.set_title("Start New Command")

        self.prefs = load_config()

        icon_pixbuf = get_icon_pixbuf("forward.png")
        if icon_pixbuf:
            self.window.set_icon(icon_pixbuf)
        self.window.set_position(Gtk.WindowPosition.CENTER)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(0)

        self.entry = Gtk.Entry()
        self.entry.set_max_length(255)
        self.entry.set_width_chars(32)
        self.entry.connect('activate', self.run_command)
        if self.xdg_menu:
            # or use menus if we have xdg data:
            hbox = Gtk.HBox(homogeneous=False, spacing=20)
            vbox.add(hbox)
            hbox.add(label("Category:"))
            self.category_combo = Gtk.ComboBoxText()
            hbox.add(self.category_combo)
            index = 0
            for i, name in enumerate(sorted(self.xdg_menu.keys())):
                self.category_combo.append_text(name)
                if name == self.prefs.get("category", ""):
                    index = i
            self.category_combo.set_active(index)
            self.category_combo.connect("changed", self.category_changed)

            hbox = Gtk.HBox(homogeneous=False, spacing=20)
            vbox.add(hbox)
            self.command_combo = Gtk.ComboBoxText()
            hbox.pack_start(label("Command:"))
            hbox.pack_start(self.command_combo)
            self.command_combo.connect("changed", self.command_changed)
            # this will populate the command combo:
            self.category_changed()
        # always show the command as text so that it can be edited:
        entry_label = label("Command to run:", font="sans 14")
        entry_al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0)
        entry_al.add(entry_label)
        vbox.add(entry_al)
        # Actual command:
        vbox.add(self.entry)

        if can_share:
            self.share = Gtk.CheckButton(label="Shared", use_underline=False)
            # shared commands will also be shown to other clients
            self.share.set_active(True)
            vbox.add(self.share)
        else:
            self.share = None

        # Buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox)
        hbox.pack_start(btn("Run", "Run this command", self.run_command, "forward.png"))
        hbox.pack_start(btn("Cancel", "", self.close, "quit.png"))

        def accel_close(*_args) -> None:
            self.close()

        add_close_accel(self.window, accel_close)
        vbox.show_all()
        self.window.vbox = vbox
        self.window.add(vbox)

    def category_changed(self, *args) -> None:
        category = self.category_combo.get_active_text()
        update_config("category", category)
        entries = typedict(typedict(self.xdg_menu.dictget(category, {})).dictget("Entries", {}))
        log("category_changed(%s) category=%s, entries=%s", args, category, entries)
        self.command_combo.get_model().clear()
        index = -1
        for i, name in enumerate(entries.keys()):
            self.command_combo.append_text(name)
            if name == self.prefs.get("command", ""):
                index = i
        if index >= 0:
            self.command_combo.set_active(index)

    def command_changed(self, *args) -> None:
        if not self.entry:
            return
        category = self.category_combo.get_active_text()
        entries = typedict(typedict(self.xdg_menu.dictget(category, {})).dictget("Entries", {}))
        command_name = self.command_combo.get_active_text()
        if command_name:
            update_config("command", command_name)
        log("command_changed(%s) category=%s, entries=%s, command_name=%s", args, category, entries, command_name)
        command = ""
        if entries and command_name:
            command_props = typedict(typedict(entries).dictget(command_name, {}))
            log("command properties=%s", command_props)
            command = typedict(command_props).strget("command")
        self.entry.set_text(command)

    def show(self) -> None:
        log("show()")
        self.window.show()
        self.window.present()

    def hide(self) -> None:
        log("hide()")
        if self.window:
            self.window.hide()

    def close(self, *args) -> bool:
        log("close%s", args)
        self.hide()
        return True

    def destroy(self, *args) -> None:
        log("close%s", args)
        if self.window:
            self.window.destroy()
            self.window = None

    def run(self) -> ExitValue:
        log("run()")
        Gtk.main()
        log("run() Gtk.main done")
        return 0

    def quit(self, *args) -> None:
        log("quit%s", args)
        self.close()
        Gtk.main_quit()

    def run_command(self, *_args) -> None:
        self.hide()
        command = self.entry.get_text()
        log("command=%s", command)
        if self.run_callback and command:
            self.run_callback(command, self.share is None or self.share.get_active())


def main() -> int:  # pragma: no cover
    from xpra.platform.gui import init as gui_init, ready as gui_ready
    from xpra.platform import program_context
    gui_init()
    with program_context("Start-New-Command", "Start New Command"):
        if "-v" in sys.argv:
            enable_debug_for("util")

        app = StartNewCommand()
        app.hide = app.quit
        register_os_signals(app.quit)
        try:
            gui_ready()
            app.show()
            return app.run()
        except KeyboardInterrupt:
            return 128 + int(signal.SIGINT)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import shlex
import inspect
import os.path
import tempfile
from collections.abc import Callable

from xpra.platform import platform_import


def valid_dir(path) -> bool:
    try:
        return bool(path) and os.path.exists(path) and os.path.isdir(path)
    except TypeError:
        return False


# helpers to easily override using env vars:
def envaslist_or_delegate(env_name: str, impl: Callable, *args):
    env_value = os.environ.get(env_name)
    if env_value is not None:
        return [env_value]
    return impl(*args)


def env_or_delegate(env_name: str, impl: Callable, *args):
    env_value = os.environ.get(env_name)
    if env_value is not None:
        return env_value
    return impl(*args)


def get_install_prefix() -> str:
    return env_or_delegate("XPRA_INSTALL_PREFIX", do_get_install_prefix)


def do_get_install_prefix() -> str:
    return sys.prefix


def get_system_conf_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SYSTEM_CONF_DIRS", do_get_system_conf_dirs)


def do_get_system_conf_dirs() -> list[str]:
    # overridden in all platforms
    return []


def get_system_menu_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SYSTEM_MENU_DIRS", do_get_system_menu_dirs)


def do_get_system_menu_dirs() -> list[str]:
    return []


def get_ssl_cert_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SSL_CERT_PATHS", do_get_ssl_cert_dirs)


def do_get_ssl_cert_dirs() -> list[str]:
    dirs = ["/etc/xpra/ssl", "/usr/local/etc/xpra/ssl", "/etc/xpra/", "/usr/local/etc/xpra"]
    if os.name != "posix" or os.getuid() != 0:
        dirs = ["~/.config/xpra/ssl", "~/.xpra/ssl"] + dirs + ["./"]
    return dirs


def get_ssl_hosts_config_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SSL_HOSTS_CONFIG_DIRS", do_get_ssl_hosts_config_dirs)


def do_get_ssl_hosts_config_dirs() -> list[str]:
    dirs = []
    for d in get_ssl_cert_dirs():
        if d.rstrip("/\\").endswith("ssl"):
            dirs.append(os.path.join(d, "hosts"))
    return dirs


def get_ssh_conf_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SSH_CONF_DIRS", do_get_ssh_conf_dirs)


def do_get_ssh_conf_dirs() -> list[str]:
    return ["/etc/ssh", "/usr/local/etc/ssh", "~/.ssh", "~/ssh"]


def get_ssh_known_hosts_files() -> list[str]:
    return envaslist_or_delegate("XPRA_SSH_KNOWN_HOSTS", do_get_ssh_known_hosts_files)


def do_get_ssh_known_hosts_files() -> list[str]:
    return ["~/.ssh/known_hosts", "~/ssh/known_hosts"]


def get_user_conf_dirs(uid=None) -> list[str]:
    return envaslist_or_delegate("XPRA_USER_CONF_DIRS", do_get_user_conf_dirs, uid)


def do_get_user_conf_dirs(_uid) -> list[str]:
    return []


def get_default_conf_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_DEFAULT_CONF_DIRS", do_get_default_conf_dirs)


def do_get_default_conf_dirs() -> list[str]:
    # some platforms may also ship a default config with the application
    return []


def get_state_dir() -> str:
    return env_or_delegate("XPRA_STATE_DIR", do_get_state_dir)


def do_get_state_dir() -> str:
    d = "~/.local/state"
    if not os.path.exists(os.path.expanduser(d)):
        return "~"
    return d


def get_sessions_dir() -> str:
    return env_or_delegate("XPRA_SESSIONS_DIR", do_get_sessions_dir)


def do_get_sessions_dir() -> str:
    return "$XDG_RUNTIME_DIR/xpra"


def get_socket_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SOCKET_DIRS", do_get_socket_dirs)


def do_get_socket_dirs() -> list[str]:
    return ["~/.xpra"]


def get_client_socket_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_CLIENT_SOCKET_DIRS", do_get_client_socket_dirs)


def do_get_client_socket_dirs() -> list[str]:
    return []


def get_default_log_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_LOG_DIRS", do_get_default_log_dirs)


def do_get_default_log_dirs() -> list[str]:
    return ["~/.xpra"]


def get_download_dir() -> str:
    return env_or_delegate("XPRA_DOWNLOAD_DIR", do_get_download_dir)


def do_get_download_dir() -> str:
    d = "~/Downloads"
    if not os.path.exists(os.path.expanduser(d)):
        return "~"
    return d


def get_mmap_dir() -> str:
    return env_or_delegate("XPRA_MMAP_DIR", do_get_mmap_dir)


def do_get_mmap_dir() -> str:
    return tempfile.gettempdir()


def get_xpra_tmp_dir() -> str:
    return env_or_delegate("XPRA_TMP_DIR", do_get_xpra_tmp_dir)


def do_get_xpra_tmp_dir() -> str:
    return tempfile.gettempdir()


def get_script_bin_dirs() -> list[str]:
    return envaslist_or_delegate("XPRA_SCRIPT_BIN_DIRS", do_get_script_bin_dirs)


def do_get_script_bin_dirs() -> list[str]:
    return ["~/.xpra"]


def get_remote_run_xpra_scripts() -> list[str]:
    return envaslist_or_delegate("XPRA_REMOTE_RUN_XPRA_SCRIPTS", do_get_remote_run_xpra_scripts)


def do_get_remote_run_xpra_scripts() -> list[str]:
    return ["xpra", "$XDG_RUNTIME_DIR/xpra/run-xpra", "/usr/local/bin/xpra", "~/.xpra/run-xpra", "Xpra_cmd.exe"]


def get_sshpass_command() -> str:
    return env_or_delegate("XPRA_SSHPASS", do_get_sshpass_command)


def do_get_sshpass_command() -> str:
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    SSHPASS = "sshpass"
    from xpra.platform.features import EXECUTABLE_EXTENSION  # pylint: disable=import-outside-toplevel
    if EXECUTABLE_EXTENSION:
        SSHPASS = f"sshpass.{EXECUTABLE_EXTENSION}"
    paths = os.environ["PATH"].split(os.pathsep)
    for path in paths:
        path = path.strip('"')
        exe_file = os.path.join(path, SSHPASS)
        if is_exe(exe_file):
            return exe_file
    return ""


# overridden in platform code:
def get_app_dir() -> str:
    return env_or_delegate("XPRA_APP_DIR", do_get_app_dir)


def do_get_app_dir() -> str:
    return default_get_app_dir()


def default_get_app_dir() -> str:
    if os.name == "posix":
        for prefix in (
                os.environ.get("RPM_BUILD_ROOT"),
                get_install_prefix(),
                sys.exec_prefix,
                "/usr",
                "/usr/local",
        ):
            if not prefix:
                continue
            adir = os.path.join(prefix, "share", "xpra")
            if valid_dir(adir):
                return adir
    adir = os.path.dirname(inspect.getfile(sys._getframe(1)))  # pylint: disable=protected-access

    def root_module(d):
        for psep in (os.path.sep, "/", "\\"):
            pos = d.find(f"xpra{psep}platform")
            if pos >= 0:
                return d[:pos]
        return d

    if valid_dir(adir):
        return root_module(adir)
    adir = os.path.dirname(sys.argv[0])
    if valid_dir(adir):
        return root_module(adir)
    adir = os.getcwd()
    return adir  # tried our best, hope this works!


# may be overridden in platform code:
def get_resources_dir() -> str:
    return env_or_delegate("XPRA_RESOURCES_DIR", do_get_resources_dir)


def do_get_resources_dir() -> str:
    return get_app_dir()


# may be overridden in platform code:

def get_image(name: str):
    filename = os.path.join(get_image_dir(), name)
    from xpra.gtk.pixbuf import get_icon_from_file
    return get_icon_from_file(filename)


def get_image_dir() -> str:
    return env_or_delegate("XPRA_IMAGE_DIR", do_get_image_dir)


def do_get_image_dir() -> str:
    raise NotImplementedError()


def get_icon_dir() -> str:
    return env_or_delegate("XPRA_ICON_DIR", do_get_icon_dir)


def do_get_icon_dir() -> str:
    raise NotImplementedError()


def get_icon(name: str):
    filename = get_icon_filename(name)
    if not filename:
        return None
    from xpra.gtk.pixbuf import get_icon_from_file
    return get_icon_from_file(filename)


def get_icon_filename(basename: str = "", ext="png") -> str:
    if not basename:
        return ""
    filename = basename
    fext = os.path.splitext(filename)[1]
    if not fext:
        filename = f"{basename}.{ext}"
    if not os.path.isabs(filename):
        icon_dir = get_icon_dir()
        filename = os.path.join(icon_dir, filename)
    if not os.path.exists(filename):
        return ""
    return os.path.abspath(filename)


def get_desktop_background_paths() -> list[str]:
    return envaslist_or_delegate("XPRA_DESKTOP_BACKGROUND_PATHS", do_get_desktop_background_paths)


def do_get_desktop_background_paths() -> list[str]:
    return []


def get_xpra_command() -> list[str]:
    envvalue = os.environ.get("XPRA_COMMAND")
    if envvalue:
        return shlex.split(envvalue)
    return do_get_xpra_command()


def do_get_xpra_command() -> list[str]:
    return default_do_get_xpra_command()


def default_do_get_xpra_command() -> list[str]:
    # try to use the same "xpra" executable that launched this server:
    if sys.argv and sys.argv[0].lower().endswith("/xpra"):
        return [sys.argv[0]]
    return ["xpra"]


def get_nodock_command() -> list[str]:
    envvalue = os.environ.get("XPRA_NODOCK_COMMAND")
    if envvalue:
        return shlex.split(envvalue)
    return do_get_nodock_command()


def do_get_nodock_command() -> list[str]:
    return get_xpra_command()


def get_audio_command() -> list[str]:
    envvalue = os.environ.get("XPRA_AUDIO_COMMAND")
    if envvalue:
        return shlex.split(envvalue)
    return do_get_audio_command()


def do_get_audio_command() -> list[str]:
    return get_xpra_command()


def get_python_exec_command() -> list[str]:
    envvalue = os.environ.get("XPRA_PYTHON_EXEC_COMMAND") or os.environ.get("XPRA_PYTHON_COMMAND")
    if envvalue:
        return shlex.split(envvalue)
    return do_get_python_exec_command()


def do_get_python_exec_command() -> list[str]:
    vi = sys.version_info
    return [f"python{vi.major}.{vi.minor}", "-c"]


def get_python_execfile_command() -> list[str]:
    envvalue = os.environ.get("XPRA_PYTHON_EXECFILE_COMMAND") or os.environ.get("XPRA_PYTHON_COMMAND")
    if envvalue:
        return shlex.split(envvalue)
    return do_get_python_execfile_command()


def do_get_python_execfile_command() -> list[str]:
    vi = sys.version_info
    return [f"python{vi.major}.{vi.minor}"]


platform_import(globals(), "paths", True,
                "do_get_resources_dir",
                "do_get_app_dir",
                "do_get_icon_dir",
                "do_get_image_dir",
                )
platform_import(globals(), "paths", False,
                "do_get_sshpass_command",
                "do_get_xpra_command",
                "do_get_audio_command",
                "do_get_nodock_command",
                "do_get_install_prefix",
                "do_get_default_conf_dirs",
                "do_get_system_conf_dirs",
                "do_get_system_menu_dirs",
                "do_get_ssl_cert_dirs",
                "do_get_ssl_hosts_config_dirs",
                "do_get_ssh_conf_dirs",
                "do_get_ssh_known_hosts_files",
                "do_get_user_conf_dirs",
                "do_get_state_dir",
                "do_get_sessions_dir",
                "do_get_socket_dirs",
                "do_get_client_socket_dirs",
                "do_get_default_log_dirs",
                "do_get_download_dir",
                "do_get_mmap_dir",
                "do_get_xpra_tmp_dir",
                "do_get_script_bin_dirs",
                "do_get_desktop_background_paths",
                "do_get_python_exec_command",
                "do_get_python_execfile_command",
                )


def get_info():
    try:
        import xpra  # pylint: disable=import-outside-toplevel
        XPRA_MODULE_PATH = xpra.__file__
        pos = XPRA_MODULE_PATH.find("__init__.py")
        if pos > 0:
            XPRA_MODULE_PATH = XPRA_MODULE_PATH[:pos]
    except AttributeError:
        XPRA_MODULE_PATH = ""
    return {
        "install": {"prefix": get_install_prefix()},
        "default_conf": {"dirs": get_default_conf_dirs()},
        "system_conf": {"dirs": get_system_conf_dirs()},
        "system-menu": {"dirs" : get_system_menu_dirs()},
        "ssl-cert": {"dirs": get_ssl_cert_dirs()},
        "ssl-hosts-config": {"dirs": get_ssl_hosts_config_dirs()},
        "ssh_conf": {"dirs": get_ssh_conf_dirs()},
        "user_conf": {"dirs": get_user_conf_dirs()},
        "state": {"dir": get_state_dir()},
        "sessions": {"dir": get_sessions_dir()},
        "socket": {"dirs": get_socket_dirs()},
        "client-socket": {"dirs": get_client_socket_dirs()},
        "log": {"dirs": get_default_log_dirs()},
        "download": {"dir": get_download_dir()},
        "mmap": {"dir": get_mmap_dir()},
        "xpra-tmp": {"dir": get_xpra_tmp_dir()},
        "script": {"dir": get_script_bin_dirs()},
        "xpra-module": XPRA_MODULE_PATH,
        "app": {"default": {"dir": default_get_app_dir()}},
        "desktop-background": get_desktop_background_paths(),
        "ssh-known-hosts": get_ssh_known_hosts_files(),
        "resources": get_resources_dir(),
        "icons": get_icon_dir(),
        "images": get_image_dir(),
        "home": os.path.expanduser("~"),
        "xpra_command": get_xpra_command(),
        "nodock_command": get_nodock_command(),
        "audio_command": get_audio_command(),
        "sshpass_command": get_sshpass_command(),
        "python-exec": get_python_exec_command(),
        "python-execfile": get_python_execfile_command(),
    }


def main():
    from xpra.log import consume_verbose_argv
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("Path-Info", "Path Info"):
        consume_verbose_argv(sys.argv, "util")
        print_nested_dict(get_info())


if __name__ == "__main__":
    main()
